"""Pipeline schedule create command implementation."""

import json
from typing import Annotated, Any
from uuid import UUID

import typer

from alloy_runtime_types.dtos.schedule import (
    CreateScheduleRequest,
    CronSchedule,
)
from cli.commands.flag_utils import validate_uuid
from cli.commands.pipelines.schedules.create_presenter import present_create_result
from cli.infrastructure.command import async_command, authenticated_client


def pipelines_schedules_create_command(
    pipeline_id: str = typer.Argument(
        ...,
        help="Pipeline UUID",
    ),
    name: str = typer.Option(
        ...,
        "-n",
        "--name",
        help="Schedule name",
    ),
    cron: str = typer.Option(
        ...,
        "--cron",
        help="Cron expression (5 fields: minute hour day month weekday)",
    ),
    description: Annotated[
        str | None,
        typer.Option(
            "-d",
            "--description",
            help="Schedule description",
        ),
    ] = None,
    timezone: str = typer.Option(
        "UTC",
        "--timezone",
        "-tz",
        help="IANA timezone (e.g. America/Chicago, Europe/London)",
    ),
    input_params: Annotated[
        str | None,
        typer.Option(
            "--input-params",
            "-i",
            help="Input parameters as JSON string",
        ),
    ] = None,
    timeout: int = typer.Option(
        300,
        "--timeout",
        "-t",
        help="Execution timeout in seconds (1-3600)",
        min=1,
        max=3600,
    ),
    disabled: bool = typer.Option(
        False,
        "--disabled",
        help="Create schedule in disabled state",
    ),
    allow_concurrent: bool = typer.Option(
        False,
        "--allow-concurrent",
        help="Allow concurrent executions",
    ),
    max_failures: int = typer.Option(
        5,
        "--max-failures",
        help="Auto-disable after this many consecutive failures (0 = never)",
        min=0,
    ),
) -> None:
    """Create a recurring schedule for a pipeline.

    Examples:
        alloy pipelines schedules create 123e4567... -n 'Daily Report' --cron '0 9 * * *'
        alloy pipelines schedules create 123e4567... -n 'Weekly' --cron '0 9 * * 1' --timezone America/Chicago
        alloy pipelines schedules create 123e4567... -n 'Hourly' --cron '0 * * * *' -i '{"key": "value"}'
    """
    p_uuid = validate_uuid(pipeline_id, "pipeline")

    params: dict[str, Any] = {}
    if input_params:
        try:
            params = json.loads(input_params)
        except json.JSONDecodeError as e:
            raise typer.BadParameter(f"Invalid JSON for --input-params: {e}")

    schedule_expr = CronSchedule(expression=cron, timezone=timezone)

    request = CreateScheduleRequest(
        name=name,
        description=description,
        schedule=schedule_expr,
        input_parameters=params,
        timeout_seconds=timeout,
        is_enabled=not disabled,
        allow_concurrent=allow_concurrent,
        max_consecutive_failures=max_failures,
    )

    _execute_create(pipeline_id=p_uuid, request=request)


@async_command
async def _execute_create(pipeline_id: UUID, request: CreateScheduleRequest) -> None:
    """Execute create schedule operation."""
    async with authenticated_client() as (_config, client):
        response = await client.create_pipeline_schedule(pipeline_id, request)

    present_create_result(response)



"""Knowledge recover command implementation."""

from typing import Optional

import typer

from cli.commands.knowledge.recover.presenter import present_recovery_result
from cli.infrastructure.command import async_command, authenticated_client
from alloy_runtime_types.dtos.knowledge import RecoverDocumentsRequest


def knowledge_recover_command(
    max_retries: Optional[int] = typer.Option(
        None,
        "-r",
        "--max-retries",
        min=1,
        max=10,
        help="Override max retry attempts (default: 3)",
    ),
    restart_queued: bool = typer.Option(
        False,
        "--restart-queued/--no-restart-queued",
        help="Also restart queued backlog documents from scratch (default: no)",
    ),
) -> None:
    """Recover stuck documents.

    Retries or permanently fails documents stuck in 'processing' state after
    their heartbeat goes stale, and logs queued backlog on the server.

    Documents can get stuck when:
    - Processing is interrupted (container crash, API timeout, etc.)
    - Celery never picks up the task (message lost, worker unavailable)

    Examples:
        alloy knowledge recover
        alloy knowledge recover --max-retries 5
        alloy knowledge recover --restart-queued
    """
    _execute_recover(max_retries=max_retries, restart_queued=restart_queued)


@async_command
async def _execute_recover(max_retries: Optional[int], restart_queued: bool) -> None:
    """Execute document recovery operation."""
    request = None
    if max_retries is not None or restart_queued:
        request = RecoverDocumentsRequest(
            max_retries=max_retries,
            restart_queued=restart_queued,
        )

    async with authenticated_client() as (_config, client):
        response = await client.recover_documents(request)

    present_recovery_result(response)



"""Schema creation modal for creating new validation schemas.

Provides a comprehensive modal screen that allows users to create a new schema
with support for JSON Schema or Regex formats, including JSON example generation.
"""

import json
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.widgets import (
    Button,
    Input,
    Label,
    RadioButton,
    RadioSet,
    Static,
    TextArea,
)

from cli.infrastructure.scope_utils import get_scope_from_radioset
from cli.tui.widgets.base_form_modal import BaseFormModal

from alloy_runtime_sdk.api_client.client import ApiClient
from alloy_runtime_types.dtos.schemas import CreateSchemaRequest
from alloy_runtime_types.enums.ownership_scope import OwnershipScope
from alloy_runtime_types.enums.schema_enums import SchemaFormat

from cli.infrastructure.forms.json_schema_builder import JsonSchemaBuilder

if TYPE_CHECKING:
    from cli.tui.app import AlloyRuntimeApp


@dataclass
class SchemaCreateResult:
    """Result of successful schema creation."""

    schema_id: str
    schema_name: str
    version: int


class SchemaCreateModal(BaseFormModal[SchemaCreateResult]):
    """Modal for creating a new schema.

    Provides a comprehensive form with:
    - Required fields: name, schema format, definition
    - Schema definition via TextArea (paste JSON schema or regex)
    - JSON Schema generation from JSON example
    - Optional fields: description
    - Scope selection (user/organization/global)

    Usage:
        def on_schema_created(result: SchemaCreateResult | None):
            if result is not None:
                # Schema was created successfully
                self.notify(f"Created schema: {result.schema_name}")
                self._do_search("")  # Refresh list

        client = await self.app.store.get_client()
        self.app.push_screen(
            SchemaCreateModal(client=client),
            on_schema_created
        )
    """

    app: "AlloyRuntimeApp"

    def __init__(
        self,
        client: ApiClient,
        **kwargs: Any,
    ) -> None:
        """Initialize the schema creation modal.

        Args:
            client: ApiClient instance for API calls
        """
        super().__init__(**kwargs)
        self._client = client

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="schema-create-modal"):
            yield Label("Create New Schema", classes="modal-title")

            # ===== REQUIRED FIELDS =====
            yield Static("[bold]Required[/]", classes="form-section")

            yield Static(
                "Name (lowercase, dots/dashes/underscores)", classes="field-label"
            )
            yield Input(
                placeholder="e.g., user.profile, api.response-format",
                id="name-input",
            )

            yield Static("Schema Format", classes="field-label")
            with RadioSet(id="format-radios"):
                yield RadioButton("JSON Schema", id="format-json-schema", value=True)
                yield RadioButton("Regex", id="format-regex")

            yield Static("Scope", classes="field-label")
            with RadioSet(id="scope-radios"):
                yield RadioButton("User (personal)", id="scope-user")
                yield RadioButton("Organization", id="scope-org", value=True)
                yield RadioButton("Global", id="scope-global")

            # ===== SCHEMA DEFINITION =====
            yield Static("[bold]Schema Definition[/]", classes="form-section")

            yield Static(
                "Definition (paste JSON Schema or Regex pattern)", classes="field-label"
            )
            yield TextArea(
                id="definition-textarea",
                classes="definition-input",
            )

            # JSON Schema builder section (only visible for JSON Schema format)
            with Vertical(id="json-builder-section"):
                yield Static(
                    "Or: Generate from JSON example (JSON Schema only)",
                    classes="field-label",
                )
                yield JsonSchemaBuilder(id="schema-builder")
                yield Button(
                    "Use Generated Schema",
                    id="use-generated-btn",
                    variant="primary",
                    disabled=True,
                )

            # ===== OPTIONAL FIELDS =====
            yield Static("[bold]Optional[/]", classes="form-section")

            yield Static("Description", classes="field-label")
            yield Input(
                placeholder="Brief description of this schema's purpose",
                id="description-input",
            )

            # Error display
            yield Static("", id="error-display", classes="error-display")

            with Horizontal(classes="button-row"):
                yield Button("Create Schema", id="create-btn", variant="success")
                yield Button("Cancel", id="cancel-btn", variant="default")

    def on_mount(self) -> None:
        """Focus the name input on mount."""
        self.query_one("#name-input", Input).focus()

    @on(RadioSet.Changed, "#format-radios")
    def on_format_changed(self, event: RadioSet.Changed) -> None:
        """Handle schema format change - show/hide JSON builder."""
        json_builder_section = self.query_one("#json-builder-section", Vertical)
        if event.pressed.id == "format-json-schema":
            json_builder_section.display = True
        else:
            json_builder_section.display = False
            # Clear the builder when switching to regex
            self.query_one("#schema-builder", JsonSchemaBuilder).clear()

    @on(JsonSchemaBuilder.SchemaChanged)
    def on_schema_builder_changed(self, event: JsonSchemaBuilder.SchemaChanged) -> None:
        """Enable/disable the 'Use Generated Schema' button based on builder state."""
        use_btn = self.query_one("#use-generated-btn", Button)
        use_btn.disabled = event.schema is None

    @on(Button.Pressed, "#use-generated-btn")
    def on_use_generated_pressed(self, event: Button.Pressed) -> None:
        """Copy the generated schema to the definition textarea."""
        builder = self.query_one("#schema-builder", JsonSchemaBuilder)
        if builder.generated_schema:
            schema_str = json.dumps(builder.generated_schema, indent=2)
            self.query_one("#definition-textarea", TextArea).text = schema_str
            self.app.notify(
                "Generated schema copied to definition", severity="information"
            )

    @on(Button.Pressed, "#create-btn")
    def on_create_pressed(self, event: Button.Pressed) -> None:
        """Handle create button press."""
        self.action_submit()

    @on(Button.Pressed, "#cancel-btn")
    def on_cancel_pressed(self, event: Button.Pressed) -> None:
        """Handle cancel button press."""
        self.dismiss(None)

    def action_submit(self) -> None:
        """Handle form submission."""
        try:
            self._validate_fields()
            self._hide_error()
            self._finalize_submission()
        except ValueError as e:
            self._show_error(str(e))

    def _validate_fields(self) -> None:
        """Validate all required fields."""
        name = self.query_one("#name-input", Input).value.strip()
        if not name:
            raise ValueError("Name is required")

        if not re.match(r"^[a-z0-9._-]+$", name):
            raise ValueError(
                "Name must be lowercase alphanumeric with dots, dashes, or underscores only"
            )

        definition = self.query_one("#definition-textarea", TextArea).text.strip()
        if not definition:
            raise ValueError(
                "Schema definition is required. Paste a JSON Schema or regex pattern."
            )

        # Validate JSON schema format
        schema_format = self._get_schema_format()
        if schema_format == SchemaFormat.JSON_SCHEMA:
            try:
                json.loads(definition)
            except json.JSONDecodeError as e:
                raise ValueError(f"Invalid JSON in schema definition: {e.msg}")

    def _get_schema_format(self) -> SchemaFormat:
        """Get schema format from RadioSet."""
        radios = self.query_one("#format-radios", RadioSet)
        pressed = radios.pressed_button
        if pressed is None or pressed.id == "format-json-schema":
            return SchemaFormat.JSON_SCHEMA
        return SchemaFormat.REGEX

    def _get_scope(self) -> OwnershipScope:
        """Get scope from RadioSet."""
        return get_scope_from_radioset(self.query_one("#scope-radios", RadioSet))

    def _build_request(self) -> CreateSchemaRequest:
        """Build the API request."""
        name = self.query_one("#name-input", Input).value.strip()
        definition = self.query_one("#definition-textarea", TextArea).text.strip()
        description = self.query_one("#description-input", Input).value.strip() or None

        return CreateSchemaRequest(
            name=name,
            schema_format=self._get_schema_format(),
            schema_definition=definition,
            description=description,
            scope=self._get_scope(),
        )

    @work(exclusive=True)
    async def _finalize_submission(self) -> None:
        """Submit to API."""
        try:
            request = self._build_request()
            response = await self._client.create_schema(request)

            result = SchemaCreateResult(
                schema_id=str(response.id),
                schema_name=response.schema_name,
                version=response.version,
            )
            self.dismiss(result)

        except Exception as e:
            self._show_error(f"Failed to create schema: {e}")



import json
from typing import Any, cast
from uuid import UUID

from alloy_runtime_types.dtos.knowledge import GetSynthesisResponse
from cli.infrastructure.formatting.fields import format_datetime
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig


def _format_optional_uuid(value: UUID | None) -> Any:
    if value is None:
        return cast(str, None)
    return str(value)


def _format_metadata(value: dict[str, Any] | None) -> Any:
    if value is None:
        return cast(str, None)
    return json.dumps(value, indent=2, sort_keys=True)


def _format_collection_ids(collection_ids: list[UUID]) -> str:
    if not collection_ids:
        return "[]"
    return "\n".join(str(collection_id) for collection_id in collection_ids)


def _format_optional_datetime(value: Any) -> Any:
    if value is None:
        return cast(str, None)
    return format_datetime(value)


def _queue_metadata_value(response: GetSynthesisResponse, key: str) -> Any:
    queue_metadata = response.queue_metadata
    if isinstance(queue_metadata, dict):
        queue_metadata_dict = cast(dict[str, Any], queue_metadata)
        return queue_metadata_dict.get(key)
    return getattr(queue_metadata, key)


def _format_request_fingerprint(response: GetSynthesisResponse) -> str:
    return cast(str, _queue_metadata_value(response, "request_fingerprint"))


def _has_attempt(response: GetSynthesisResponse) -> bool:
    attempt_number = _queue_metadata_value(response, "attempt_number")
    total_attempt_count = _queue_metadata_value(response, "total_attempt_count")
    return attempt_number is not None and total_attempt_count is not None


def _format_attempt(response: GetSynthesisResponse) -> str:
    attempt_number = _queue_metadata_value(response, "attempt_number")
    total_attempt_count = _queue_metadata_value(response, "total_attempt_count")
    return f"{attempt_number} of {total_attempt_count}"


def _has_failure_classification(response: GetSynthesisResponse) -> bool:
    return _queue_metadata_value(response, "failure_classification") is not None


def _format_failure_classification(response: GetSynthesisResponse) -> str:
    return cast(str, _queue_metadata_value(response, "failure_classification"))


def present_synthesis_details(response: GetSynthesisResponse) -> None:
    output = OutputService.get()

    fields = [
        FieldConfig("id", "ID", str),
        FieldConfig("title", "Title"),
        FieldConfig("topic", "Topic"),
        FieldConfig("status", "Status"),
        FieldConfig("is_stale", "Is Stale"),
        FieldConfig("staleness_reason", "Staleness Reason"),
        FieldConfig("source_chunk_count", "Source Chunks"),
        FieldConfig("collection_ids", "Collections", _format_collection_ids),
        FieldConfig(
            "synthesizer_agent_id",
            "Synthesizer Agent ID",
            _format_optional_uuid,
        ),
        FieldConfig("execution_id", "Execution ID", _format_optional_uuid),
        FieldConfig("error_message", "Error Message"),
        FieldConfig("metadata", "Metadata", _format_metadata),
        FieldConfig("synthesis_content", "Synthesis Content"),
        FieldConfig("created_at", "Created At", _format_optional_datetime),
        FieldConfig("updated_at", "Updated At", _format_optional_datetime),
    ]

    if response.queue_metadata is not None:
        queue_fields = [
            FieldConfig(
                "queue_metadata",
                "Request Fingerprint",
                lambda _value: _format_request_fingerprint(response),
            ),
        ]
        if _has_attempt(response):
            queue_fields.append(
                FieldConfig(
                    "queue_metadata",
                    "Attempt",
                    lambda _value: _format_attempt(response),
                )
            )
        if _has_failure_classification(response):
            queue_fields.append(
                FieldConfig(
                    "queue_metadata",
                    "Failure Classification",
                    lambda _value: _format_failure_classification(response),
                )
            )
        fields[10:10] = queue_fields

    output.entity(
        response,
        f"Synthesis: {response.title}",
        fields,
        raw_payload=response,
    )



"""Presenter for schema update output."""

from cli.infrastructure.formatting.fields import (
    format_datetime,
    format_optional,
    format_uuid,
)
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig
from alloy_runtime_types.dtos.schemas import UpdateSchemaResponse


def _derive_scope(response: UpdateSchemaResponse) -> str:
    """Derive scope string from organization_id and user_id fields."""
    if response.organization_id is not None and response.user_id is None:
        return "organization"
    elif response.user_id is not None:
        return "user"
    else:
        return "global"


def present_schema_update_success(response: UpdateSchemaResponse) -> None:
    """Present successful schema update.

    Args:
        response: Schema update response from server
    """
    output = OutputService.get()
    output.result(
        action="update",
        resource="schema",
        identifier=str(response.id),
        name=response.schema_name,
    )

    fields = [
        FieldConfig("id", "ID", lambda v: format_uuid(v, short=False)),
        FieldConfig("schema_name", "Name"),
        FieldConfig("schema_format_id", "Format"),
        FieldConfig("version", "Version", str),
        FieldConfig("description", "Description", format_optional),
        FieldConfig("scope", "Scope", lambda _: _derive_scope(response)),
        FieldConfig(
            "organization_id",
            "Organization ID",
            lambda v: format_optional(v, lambda x: format_uuid(x, short=True)),
        ),
        FieldConfig(
            "user_id",
            "User ID",
            lambda v: format_optional(v, lambda x: format_uuid(x, short=True)),
        ),
        FieldConfig("is_latest", "Is Latest", str),
        FieldConfig("created_at", "Created At", format_datetime),
    ]

    output.entity(
        response, f"Schema: {response.schema_name}", fields, raw_payload=response
    )



"""Command for undoing the last conversation turn."""

from dataclasses import dataclass

from alloy_runtime_sdk.api_client.client import ApiClient
from alloy_runtime_types.dtos.messages import UpdateMessageRequest
from alloy_runtime_types.dtos.sessions import MessageResponse

from cli.tui.chat.commands.base import CommandResult
from cli.tui.chat.types import extract_message_text


@dataclass
class UndoResult:
    """Result data for undo operation.

    Attributes:
        user_preview: Preview of the undone user message.
    """

    user_preview: str


@dataclass
class UndoCommand:
    """Command to undo the last conversation turn.

    This marks both the last user message and assistant response as
    excluded from context, effectively removing them from the conversation.

    Attributes:
        client: Server client for API calls.
        user_message: The user message to exclude.
        assistant_message: The assistant message to exclude.
    """

    client: ApiClient
    user_message: MessageResponse
    assistant_message: MessageResponse

    async def execute(self) -> CommandResult[UndoResult]:
        """Execute the undo operation.

        Returns:
            CommandResult with preview of undone message on success.
        """
        try:
            # Mark both messages as excluded from context
            for msg in [self.user_message, self.assistant_message]:
                await self.client.update_message(
                    msg.id, UpdateMessageRequest(include_in_context=False)
                )

            # Get preview text for user feedback
            user_text = extract_message_text(self.user_message)
            user_preview = user_text

            return CommandResult[UndoResult](
                success=True,
                data=UndoResult(user_preview=user_preview),
            )

        except Exception as e:
            return CommandResult[UndoResult](success=False, error=str(e))



"""Presenter for session delete command output."""

from cli.infrastructure.output import OutputService
from alloy_runtime_types.dtos.sessions import DeleteSessionResponse


def present_delete_result(response: DeleteSessionResponse) -> None:
    """Present session deletion result."""
    output = OutputService.get()
    output.result(
        action="delete",
        resource="session",
        identifier=str(response.deleted_session.id),
        name=None,
    )



"""Pipeline execution rerun command implementation."""

from uuid import UUID

import typer

from cli.commands.flag_utils import validate_uuid
from cli.commands.pipelines.execute.presenter import present_execution_result
from cli.infrastructure.command import async_command, authenticated_client


def pipelines_executions_rerun_command(
    pipeline_id: str = typer.Argument(
        ...,
        help="Pipeline UUID",
    ),
    execution_id: str = typer.Argument(
        ...,
        help="Execution UUID to rerun",
    ),
) -> None:
    """Re-run a failed or paused pipeline execution.

    Completed steps are automatically skipped via idempotent cache-hits.
    Use after approving a checkpoint to resume pipeline execution.

    Examples:
        alloy pipelines executions rerun 123e4567... 456e7890...
    """
    p_uuid = validate_uuid(pipeline_id, "pipeline")
    e_uuid = validate_uuid(execution_id, "execution")
    _execute_rerun(pipeline_id=p_uuid, execution_id=e_uuid)


@async_command
async def _execute_rerun(pipeline_id: UUID, execution_id: UUID) -> None:
    """Execute rerun pipeline operation."""
    async with authenticated_client() as (_config, client):
        response = await client.rerun_pipeline_execution(pipeline_id, execution_id)

    present_execution_result(response)



"""Presenter for knowledge search output."""

import json
from typing import Any, cast

from pydantic import BaseModel
from alloy_runtime_types.dtos.knowledge import SearchResponse
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig
from cli.infrastructure.renderers.list_renderer import ColumnConfig


def _format_optional_text(value: str | None) -> Any:
    if value is None:
        return cast(str, None)
    return value


def _format_timing(value: dict[str, int]) -> str:
    if not value:
        return "{}"
    return ", ".join(f"{key}={timing_ms}ms" for key, timing_ms in value.items())


def _format_optional_list(value: list[str] | None) -> Any:
    if not value:
        return cast(str, None)
    return "\n".join(value)


def _format_tags(value: list[Any]) -> str:
    if not value:
        return "[]"
    return ", ".join(tag.tag_path for tag in value)


def _format_metadata(value: dict[str, Any] | None) -> Any:
    if not value:
        return cast(str, None)
    return json.dumps(value, sort_keys=True, default=str)


def _format_score(value: float) -> str:
    return f"{value:.3f}"


def present_search_results(response: SearchResponse) -> None:
    output = OutputService.get()

    output.entity(
        response,
        "Search Summary",
        [
            FieldConfig("query", "Query"),
            FieldConfig("results", "Results", lambda value: str(len(value))),
            FieldConfig("total_candidates", "Candidates"),
            FieldConfig("avg_relevance_score", "Avg Relevance", _format_score),
            FieldConfig("timing_ms", "Timing", _format_timing),
            FieldConfig(
                "transformations_applied",
                "Transformations",
                lambda value: ", ".join(value) if value else cast(str, None),
            ),
            FieldConfig("sub_queries_used", "Sub Queries", _format_optional_list),
            FieldConfig("original_query", "Original Query", _format_optional_text),
            FieldConfig("search_mode", "Search Mode", _format_optional_text),
        ],
        raw_payload=response,
    )

    output.table(
        cast(list[BaseModel], response.results),
        "Search Results",
        [
            ColumnConfig("document_title", "Document"),
            ColumnConfig("relevance_score", "Score", _format_score),
            ColumnConfig("low_confidence", "Low Confidence"),
            ColumnConfig("document_tags", "Tags", _format_tags),
            ColumnConfig("document_metadata", "Metadata", _format_metadata),
            ColumnConfig("content", "Content"),
        ],
        raw_payload=response.results,
    )



"""Pipeline create command implementation."""

from pathlib import Path
from typing import Optional

import typer

from cli.commands.pipelines.create.presenter import present_pipeline_create_success
from cli.infrastructure.command import async_command, authenticated_client
from alloy_runtime_types.dtos.pipeline import CreatePipelineRequest


def pipelines_create_command(
    name: str = typer.Option(
        ...,
        "--name",
        "-n",
        help="Pipeline name",
    ),
    description: Optional[str] = typer.Option(
        None,
        "--description",
        "-d",
        help="Pipeline description",
    ),
    file: Path = typer.Option(
        ...,
        "--file",
        "-f",
        help="Python file containing @pipeline decorated function",
        exists=True,
        readable=True,
    ),
    function_name: Optional[str] = typer.Option(
        None,
        "--function-name",
        help="Name of the @pipeline function to run (uses first/only if not specified)",
    ),
    public: bool = typer.Option(
        False,
        "--public/--no-public",
        help="Make pipeline publicly accessible",
    ),
) -> None:
    """Create a new pipeline from a Python file.

    The file must contain a @pipeline decorated function.

    Examples:
        alloy pipelines create --name my-pipeline --file pipeline.py
        alloy pipelines create -n my-pipeline -f pipeline.py --public
        alloy pipelines create -n my-pipeline -f pipeline.py --function-name main_pipeline
    """
    _validate_file_extension(file)
    _execute_create(
        name=name,
        description=description,
        file=file,
        function_name=function_name,
        public=public,
    )


def _validate_file_extension(file: Path) -> None:
    """Validate file has .py extension."""
    if file.suffix != ".py":
        raise typer.BadParameter(
            f"File must have .py extension, got: {file.suffix}",
            param_hint="--file",
        )


def _execute_create(
    name: str,
    description: Optional[str],
    file: Path,
    function_name: Optional[str],
    public: bool,
) -> None:
    """Build and execute pipeline creation from command flags."""
    module_content = file.read_text()

    if not module_content.strip():
        raise typer.BadParameter("Pipeline file cannot be empty", param_hint="--file")

    request = CreatePipelineRequest(
        name=name,
        description=description,
        module_content=module_content,
        pipeline_function_name=function_name,
        is_public=public,
    )

    _create_pipeline(request)


@async_command
async def _create_pipeline(request: CreatePipelineRequest) -> None:
    """Create pipeline via API."""
    async with authenticated_client() as (_config, client):
        response = await client.create_pipeline(request)

    present_pipeline_create_success(response)



"""Pipeline get command implementation."""

from uuid import UUID

import typer

from cli.commands.flag_utils import validate_uuid
from cli.commands.pipelines.get.presenter import present_pipeline
from cli.infrastructure.command import async_command, authenticated_client


def pipelines_get_command(
    pipeline_id: str = typer.Argument(
        ...,
        help="Pipeline UUID",
    ),
) -> None:
    """Get details of a pipeline.

    Examples:
        alloy pipelines get 123e4567-89ab-cdef-0123-456789abcdef
    """
    pipeline_uuid = validate_uuid(pipeline_id, "pipeline")
    _execute_get(pipeline_id=pipeline_uuid)


@async_command
async def _execute_get(pipeline_id: UUID) -> None:
    """Execute get pipeline operation."""
    async with authenticated_client() as (_config, client):
        response = await client.get_pipeline(pipeline_id)

    present_pipeline(response)



"""Render HTML to image command implementation."""

import base64
from pathlib import Path

import typer

from cli.infrastructure.command import async_command, authenticated_client
from cli.infrastructure.file_content import FileContentError, resolve_content
from cli.infrastructure.output import OutputService
from cli.infrastructure.spinner import spinner
from alloy_runtime_types.dtos.render import RenderHtmlToImageRequest


def render_html_to_image_command(
    html: str = typer.Option(
        ...,
        "-h",
        "--html",
        help="HTML text to render (or @file to read from file)",
    ),
    output_file: str = typer.Option(
        ...,
        "-o",
        "--output",
        help="Output file path for the PNG image",
    ),
    bg_color: str = typer.Option(
        "#ffffff",
        "--bg",
        "--background",
        help="Background color in hex format (default: white)",
    ),
    text_color: str = typer.Option(
        "#000000",
        "--fg",
        "--foreground",
        help="Text color in hex format (default: black)",
    ),
    font_family: str = typer.Option(
        "Georgia, serif",
        "--font",
        help="CSS font-family value (used as fallback when --font-url is provided)",
    ),
    font_url: str | None = typer.Option(
        None,
        "--font-url",
        help="Google Fonts URL (font name is auto-extracted from URL)",
    ),
    width: int = typer.Option(
        1080,
        "-W",
        "--width",
        min=100,
        max=4096,
        help="Image width in pixels (default: 1080 for social media)",
    ),
    height: int = typer.Option(
        1080,
        "-H",
        "--height",
        min=100,
        max=4096,
        help="Image height in pixels (default: 1080 for social media)",
    ),
) -> None:
    """Render HTML text to a PNG image.

    Creates a square 1080x1080 image by default, perfect for social media ads
    (Instagram, Facebook). Uses white background with black text.

    Examples:

        # Basic usage with inline HTML
        alloy render image -h "<strong>Hello</strong> World!" -o output.png

        # Read HTML from file
        alloy render image -h @content.html -o output.png

        # Custom colors (dark mode)
        alloy render image -h "Dark mode" -o dark.png --bg "#1a1a1a" --fg "#ffffff"

        # Custom dimensions (Instagram story: 1080x1920)
        alloy render image -h "Story" -o story.png -W 1080 -H 1920

        # Custom font
        alloy render image -h "Custom" -o custom.png --font "Arial, sans-serif"

        # Google Font (font name auto-extracted from URL)
        alloy render image -h "Elegant" -o elegant.png \\
            --font-url "https://fonts.googleapis.com/css2?family=Playfair+Display:wght@700"
    """
    # Resolve @file references for HTML content
    try:
        resolved_html = resolve_content(html)
    except FileContentError as e:
        raise typer.BadParameter(str(e))

    if not resolved_html:
        raise typer.BadParameter("HTML content cannot be empty")

    # Validate hex color format
    import re

    hex_pattern = r"^#[0-9a-fA-F]{6}$"
    if not re.match(hex_pattern, bg_color):
        raise typer.BadParameter(
            f"Invalid background color '{bg_color}'. Use hex format like #ffffff"
        )
    if not re.match(hex_pattern, text_color):
        raise typer.BadParameter(
            f"Invalid text color '{text_color}'. Use hex format like #000000"
        )

    _execute_render(
        html=resolved_html,
        output_file=output_file,
        bg_color=bg_color,
        text_color=text_color,
        font_family=font_family,
        font_url=font_url,
        width=width,
        height=height,
    )


@async_command
async def _execute_render(
    html: str,
    output_file: str,
    bg_color: str,
    text_color: str,
    font_family: str,
    font_url: str | None,
    width: int,
    height: int,
) -> None:
    """Execute the render HTML to image API call."""
    output = OutputService.get()

    async with authenticated_client() as (_config, client):
        request = RenderHtmlToImageRequest(
            html_text=html,
            bg_color=bg_color,
            text_color=text_color,
            font_family=font_family,
            font_url=font_url,
            width=width,
            height=height,
        )

        async with spinner(f"Rendering {width}x{height} image..."):
            response = await client.render_html_to_image(request)

    # Decode base64 and write to file
    image_bytes = base64.b64decode(response.image_base64)
    output_path = Path(output_file)

    # Create parent directories if needed
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # Write the image
    output_path.write_bytes(image_bytes)

    file_size_kb = len(image_bytes) / 1024
    output.result(
        action="write",
        resource="image",
        identifier=None,
        name=output_path.name,
        extra_fields={
            "Path": str(output_path),
            "Width": response.width,
            "Height": response.height,
            "Size": f"{file_size_kb:.1f} KB",
            "Render Time": f"{response.render_time_ms} ms",
        },
    )



"""Simple application store for shared state.

This module provides a minimal store that holds:
- API credentials (read-only config)
- Shared ApiClient instance (lazy initialization)
- Agent cache for cross-screen sharing

This replaces the previous ELM-style architecture with Textual's
recommended "App as Store" pattern - simple and direct.
"""

from alloy_runtime_sdk.api_client.client import ApiClient
from alloy_runtime_types.dtos.agents import AgentSummary


class AppStore:
    """Simple app-wide store for shared state.

    Usage:
        store = AppStore(api_url="...", api_key="...")
        client = await store.get_client()

    Screens access via self.app.store (Textual's recommended pattern).
    """

    def __init__(self, api_url: str, api_key: str) -> None:
        """Initialize the store.

        Args:
            api_url: Base URL for the API server
            api_key: API key for authentication
        """
        self.api_url = api_url.rstrip("/")
        self.api_key = api_key
        self._client: ApiClient | None = None
        self._client_entered: bool = False

        # Cross-screen cache for agent picker in chat
        self.agent_cache: list[AgentSummary] = []

    async def get_client(self) -> ApiClient:
        """Get the shared server client instance.

        Creates and connects on first use, reuses thereafter.

        Returns:
            Connected ApiClient instance

        Raises:
            Exception: If connection fails
        """
        if self._client is None:
            self._client = ApiClient(
                base_url=self.api_url,
                api_key=self.api_key,
                timeout=30.0,
            )

        if not self._client_entered:
            await self._client.__aenter__()
            self._client_entered = True

        return self._client

    async def close(self) -> None:
        """Close the client connection.

        Should be called when the application exits.
        """
        if self._client and self._client_entered:
            await self._client.__aexit__(None, None, None)
            self._client_entered = False
            self._client = None



"""Tool configs list command implementation."""

from typing import Optional

import typer

from cli.commands.tool_configs.list.presenter import present_tool_configs_list
from cli.infrastructure.command import async_command, authenticated_client


def tool_configs_list_command(
    search: Optional[str] = typer.Option(
        None,
        "-s",
        "--search",
        help="Fuzzy search by name or description",
    ),
    limit: int = typer.Option(
        50,
        "-l",
        "--limit",
        min=1,
        max=100,
        help="Max results",
    ),
    offset: int = typer.Option(
        0,
        "--offset",
        min=0,
        help="Skip N results",
    ),
) -> None:
    """List tool configs accessible to you.

    Shows tool configs owned by you, your organization, or global configs.

    Examples:
        alloy tool-configs list
        alloy tool-configs list -s "rag"
        alloy tool-configs list -l 20
    """
    _execute_list(search=search, limit=limit, offset=offset)


@async_command
async def _execute_list(
    search: Optional[str],
    limit: int,
    offset: int,
) -> None:
    """Execute list tool configs operation."""
    async with authenticated_client() as (_config, client):
        response = await client.list_tool_configs(
            search=search,
            limit=limit,
            offset=offset,
        )

    present_tool_configs_list(response)



"""Presenter for agents get command output."""

from alloy_runtime_types.dtos.agents import GetAgentResponse

from cli.commands.agents.tool_formatting import format_tool_details
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig


def _format_configured_count(values: list[object]) -> str:
    return f"{len(values)} configured"


def _format_tags(response: GetAgentResponse) -> str:
    if not response.tags:
        return "[]"
    return ", ".join(tag.tag_path for tag in response.tags)


def present_agent_details(response: GetAgentResponse) -> None:
    output = OutputService.get()
    fields = [
        FieldConfig("id", "ID", str),
        FieldConfig("name", "Name"),
        FieldConfig("description", "Description"),
        FieldConfig("models", "Models", _format_configured_count),
        FieldConfig("tools", "Tools", lambda _: format_tool_details(response.tools)),
        FieldConfig("id", "Tags", lambda _: _format_tags(response)),
    ]

    output.entity(
        response,
        f"Agent: {response.name}",
        fields,
        raw_payload=response,
    )



"""Knowledge documents delete command implementation."""

import typer

from cli.commands.knowledge.documents.delete.presenter import present_delete_result
from cli.infrastructure.command import async_command, authenticated_client
from cli.infrastructure.console import Confirm


def knowledge_documents_delete_command(
    document: str = typer.Argument(
        ...,
        help="Document UUID",
    ),
    yes: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip confirmation",
    ),
) -> None:
    """Delete a document and all its chunks.

    Permanently removes the document and all associated chunks and questions.

    Examples:
        alloy knowledge documents delete 550e8400-e29b-41d4-a716-446655440000
        alloy knowledge documents delete 550e8400-e29b-41d4-a716-446655440000 -y
    """
    _execute_delete(document_id=document, skip_confirm=yes)


@async_command
async def _execute_delete(document_id: str, skip_confirm: bool) -> None:
    """Execute delete document operation."""
    if not skip_confirm:
        confirmed = Confirm.ask(
            f"Delete document '{document_id}'? "
            "This will delete all chunks and cannot be undone"
        )
        if not confirmed:
            raise typer.Abort()

    async with authenticated_client() as (_config, client):
        response = await client.delete_document(document_id)

    present_delete_result(response)



from cli.infrastructure.output import OutputService
from alloy_runtime_types.dtos.knowledge import BulkReingestFailedDocumentsResponse


def _follow_up_message(response: BulkReingestFailedDocumentsResponse) -> str | None:
    if response.dry_run:
        return None
    if response.status == "accepted":
        return f"Bulk reingest accepted. Poll status with: alloy knowledge documents reingest-failed-status {response.operation_id}"
    if response.queued_count and response.queued_count > 0:
        return "Bulk reingest queued. Watch worker queue depth until processing stabilizes."
    if response.skipped_count and response.skipped_count > 0:
        return "No documents were queued. Review skipped documents before retrying."
    return "No failed documents matched the provided scope."


def present_bulk_reingest_failed_result(
    response: BulkReingestFailedDocumentsResponse,
) -> None:
    output = OutputService.get()
    next_step = _follow_up_message(response)

    extra_fields: dict[str, bool | str | int | None] = {
        "Dry Run": response.dry_run,
        "Message": response.message,
        "Next Step": next_step,
    }

    if response.status == "accepted":
        extra_fields["Status"] = response.status
        extra_fields["Operation ID"] = str(response.operation_id)
    else:
        extra_fields["Matched"] = response.matched_count
        extra_fields["Queued"] = response.queued_count
        extra_fields["Skipped"] = response.skipped_count

    output.result(
        action="reingest",
        resource="knowledge_documents",
        status="dry_run" if response.dry_run else "success",
        identifier=str(response.operation_id) if response.operation_id else None,
        name=None,
        extra_fields=extra_fields,
    )



"""Agents get command implementation."""

import typer

from cli.commands.agents.get.presenter import present_agent_details
from cli.infrastructure.command import async_command, authenticated_client


def agents_get_command(
    agent: str = typer.Argument(
        ...,
        help="Agent name or UUID",
    ),
) -> None:
    """Get details of a single agent.

    Retrieves the full agent configuration including models and tools.
    The agent can be specified by name or UUID.

    Examples:
        alloy agents get my-agent
        alloy agents get 123e4567-89ab-cdef-0123-456789abcdef
    """
    _execute_get(agent_identifier=agent)


@async_command
async def _execute_get(agent_identifier: str) -> None:
    """Execute get agent operation.

    Args:
        agent_identifier: Agent name or UUID
    """
    async with authenticated_client() as (_config, client):
        response = await client.get_agent(agent_identifier)

    present_agent_details(response)



from collections.abc import Mapping
from typing import Any, cast

from click import Context


def build_help_json(ctx: Context) -> dict[str, Any]:
    command_info: dict[str, Any] = dict(ctx.command.to_info_dict(ctx))
    usage = " ".join([ctx.command_path, *ctx.command.collect_usage_pieces(ctx)]).strip()
    payload = _normalize_command(command_info)
    payload["usage"] = usage
    payload["description"] = payload.pop("help")
    payload.pop("deprecated", None)
    payload.pop("hidden", None)
    return payload


def _normalize_command(info: dict[str, Any]) -> dict[str, Any]:
    option_entries: list[dict[str, Any]] = []
    for raw_parameter in cast(list[Any], info.get("params", [])):
        if not isinstance(raw_parameter, Mapping):
            continue
        parameter = dict(cast(Mapping[str, Any], raw_parameter))
        if bool(parameter.get("hidden", False)):
            continue
        option_entries.append(_normalize_parameter(parameter))

    command_entries: list[dict[str, Any]] = []
    for raw_command in _iter_commands(info.get("commands", {})):
        if bool(raw_command.get("hidden", False)):
            continue
        command_entries.append(_normalize_command(raw_command))

    return {
        "name": str(info["name"]),
        "help": _normalize_help_text(info),
        "deprecated": bool(info.get("deprecated", False)),
        "hidden": bool(info.get("hidden", False)),
        "options": option_entries,
        "commands": command_entries,
    }


def _iter_commands(commands: Any) -> list[dict[str, Any]]:
    if not isinstance(commands, Mapping):
        return []
    command_mapping = cast(Mapping[Any, Any], commands)
    normalized_commands: list[dict[str, Any]] = []
    for raw_command in command_mapping.values():
        if isinstance(raw_command, Mapping):
            normalized_commands.append(dict(cast(Mapping[str, Any], raw_command)))
    return normalized_commands


def _normalize_parameter(parameter: dict[str, Any]) -> dict[str, Any]:
    parameter_type = str(parameter.get("param_type_name", "option"))
    choices = _extract_choices(parameter.get("type"))
    payload: dict[str, Any] = {
        "name": _parameter_name(parameter, parameter_type),
        "kind": parameter_type,
        "required": bool(parameter.get("required", False)),
        "default": _json_scalar(parameter.get("default")),
        "help": _normalize_help_text(parameter),
    }
    envvar = parameter.get("envvar")
    if envvar is not None:
        payload["envvar"] = str(envvar)
    if bool(parameter.get("multiple", False)):
        payload["multiple"] = True
    if choices is not None:
        payload["choices"] = choices
    return payload


def _parameter_name(parameter: dict[str, Any], parameter_type: str) -> str:
    if parameter_type == "argument":
        return str(parameter["name"])
    options = cast(list[Any], parameter.get("opts", []))
    if options:
        return str(options[0])
    return str(parameter["name"])


def _extract_choices(type_info: Any) -> list[Any] | None:
    if not isinstance(type_info, Mapping):
        return None
    type_mapping = cast(Mapping[str, Any], type_info)
    raw_choices = type_mapping.get("choices")
    if not isinstance(raw_choices, (list, tuple)):
        return None
    normalized_choices: list[Any] = []
    for choice in cast(list[Any] | tuple[Any, ...], raw_choices):
        normalized_choices.append(_json_scalar(choice))
    return normalized_choices


def _normalize_help_text(info: Mapping[str, Any]) -> str | None:
    help_text = cast(Any, info.get("help")) or cast(Any, info.get("short_help"))
    if help_text is None:
        return None
    return str(help_text)


def _json_scalar(value: Any) -> Any:
    if value is None or isinstance(value, (str, int, float, bool)):
        return value
    if isinstance(value, tuple):
        tuple_items = cast(tuple[Any, ...], value)
        return [_json_scalar(item) for item in tuple_items]
    if isinstance(value, list):
        list_items = cast(list[Any], value)
        return [_json_scalar(item) for item in list_items]
    if isinstance(value, dict):
        dict_items = cast(dict[Any, Any], value)
        return {str(key): _json_scalar(item) for key, item in dict_items.items()}
    return str(value)



"""Alloy Runtime CLI main entry point.

This module defines the root Typer application and registers all command groups.
Running `alloy` with no arguments launches the unified TUI application.
"""

import json
import errno
import os
import sys
from collections.abc import Callable, Iterator, Sequence
from contextlib import contextmanager
from enum import Enum
from importlib import import_module
from pathlib import Path
from typing import Any, Optional, TextIO, cast

import click
import typer
from click.core import ParameterSource
from rich.console import Console
from typer.main import get_command
from typer.core import TyperGroup

from cli.commands.admin.credentials.create.command import (
    admin_credentials_create_command,
)
from cli.commands.admin.credentials.grant.command import (
    admin_credentials_grant_command,
)
from cli.commands.api_keys.create.command import api_keys_create_command
from cli.commands.agents.create.command import agents_create_command
from cli.commands.agents.delete.command import agents_delete_command
from cli.commands.agents.list.command import agents_list_command
from cli.commands.agents.update.command import agents_update_command
from cli.commands.auth.login.command import login_command
from cli.commands.auth.signup.command import signup_command
from cli.commands.ping.command import ping_command
from cli.commands.agents.get.command import agents_get_command
from cli.commands.audio.transcribe.command import audio_transcribe_command
from cli.commands.billing.costs.by_agent.command import billing_costs_by_agent_command
from cli.commands.billing.costs.by_model.command import billing_costs_by_model_command
from cli.commands.billing.costs.daily.command import billing_costs_daily_command
from cli.commands.billing.costs.summary.command import billing_costs_summary_command
from cli.commands.billing.projects.create.command import billing_projects_create_command
from cli.commands.billing.projects.get.command import billing_projects_get_command
from cli.commands.billing.projects.list.command import billing_projects_list_command
from cli.commands.generate.cancel.command import generate_cancel_command
from cli.commands.generate.status.command import generate_status_command
from cli.commands.knowledge.collections.cluster.command import (
    knowledge_collections_cluster_command,
)
from cli.commands.knowledge.collections.cluster_status.command import (
    knowledge_collections_cluster_status_command,
)
from cli.commands.knowledge.collections.status.command import (
    knowledge_collections_status_command,
)
from cli.commands.knowledge.documents.bulk_metadata.command import (
    knowledge_documents_bulk_metadata_command,
)
from cli.commands.knowledge.documents.count.command import (
    knowledge_documents_count_command,
)
from cli.commands.knowledge.documents.bulk_reingest_failed.command import (
    knowledge_documents_bulk_reingest_failed_command,
)
from cli.commands.knowledge.documents.bulk_reingest_failed_status.command import (
    knowledge_documents_bulk_reingest_failed_status_command,
)
from cli.commands.knowledge.documents.reingest.command import (
    knowledge_documents_reingest_command,
)
from cli.commands.knowledge.documents.update.command import (
    knowledge_documents_update_command,
)
from cli.commands.knowledge.synthesis.update.command import (
    knowledge_synthesis_update_command,
)
from cli.commands.organizations.create.command import organizations_create_command
from cli.commands.pipelines.create.command import pipelines_create_command
from cli.commands.pipelines.env_vars.command import pipelines_env_vars_command
from cli.commands.pipelines.update.command import pipelines_update_command
from cli.commands.pipelines.executions.costs.command import (
    pipeline_execution_costs_command,
)
from cli.commands.pipelines.executions.costs_by_model.command import (
    pipeline_execution_costs_by_model_command,
)
from cli.commands.pipelines.executions.update.command import (
    pipeline_execution_update_command,
)

from cli.commands.pipelines.schedules.env_vars_command import (
    pipelines_schedules_env_vars_command,
)
from cli.commands.pipelines.schedules.create_command import (
    pipelines_schedules_create_command,
)
from cli.commands.pipelines.schedules.update_command import (
    pipelines_schedules_update_command,
)
from cli.commands.pipelines.schedules.once_command import (
    pipelines_schedules_once_command,
)
from cli.commands.schemas.delete.command import schemas_delete_command
from cli.commands.tags.get.command import tags_get_command
from cli.commands.tags.update.command import tags_update_command
from cli.commands.templates.get_by_version.command import (
    templates_get_by_version_command,
)
from cli.commands.tools.get.command import tools_get_command
from cli.commands.tools.list.command import tools_list_command
from cli.commands.users.create.command import users_create_command

from cli.commands.content.edit.command import content_edit_command
from cli.commands.content.get.command import content_get_command
from cli.commands.content.list.command import content_list_command
from cli.commands.credentials.create.command import credentials_create_command
from cli.commands.credentials.list.command import credentials_list_command
from cli.commands.credentials.supported.command import credentials_supported_command
from cli.commands.credentials.update.command import credentials_update_command
from cli.commands.generate.text.command import generate_text_command
from cli.commands.models.list.command import models_list_command
from cli.commands.schemas.create.command import schemas_create_command
from cli.commands.schemas.get.command import schemas_get_command
from cli.commands.schemas.list.command import schemas_list_command
from cli.commands.sync.command import sync_command
from cli.commands.admin.bootstrap_command import bootstrap_admin_command
from cli.commands.templates.create.command import templates_create_command
from cli.commands.templates.delete.command import templates_delete_command
from cli.commands.templates.get.command import templates_get_command
from cli.commands.templates.list.command import templates_list_command
from cli.commands.templates.render.command import templates_render_command
from cli.commands.templates.validate.command import templates_validate_command
from cli.commands.templates.update.command import templates_update_command
from cli.commands.templates.version.command import templates_version_command
from cli.commands.content.delete.command import content_delete_command
from cli.commands.executions.get.command import executions_get_command
from cli.commands.schemas.update.command import schemas_update_command
from cli.commands.render.html_to_image.command import render_html_to_image_command
from cli.commands.knowledge.collections.create.command import (
    knowledge_collections_create_command,
)
from cli.commands.knowledge.collections.delete.command import (
    knowledge_collections_delete_command,
)
from cli.commands.knowledge.collections.get.command import (
    knowledge_collections_get_command,
)
from cli.commands.knowledge.collections.list.command import (
    knowledge_collections_list_command,
)
from cli.commands.knowledge.collections.update.command import (
    knowledge_collections_update_command,
)
from cli.commands.knowledge.documents.delete.command import (
    knowledge_documents_delete_command,
)
from cli.commands.knowledge.documents.get.command import (
    knowledge_documents_get_command,
)
from cli.commands.knowledge.documents.ingest.command import (
    knowledge_documents_ingest_command,
)
from cli.commands.knowledge.documents.list.command import (
    knowledge_documents_list_command,
)
from cli.commands.knowledge.recover.command import knowledge_recover_command
from cli.commands.knowledge.search.command import knowledge_search_command
from cli.commands.knowledge.synthesis.create.command import (
    knowledge_synthesis_create_command,
)
from cli.commands.knowledge.synthesis.delete.command import (
    knowledge_synthesis_delete_command,
)
from cli.commands.knowledge.synthesis.get.command import (
    knowledge_synthesis_get_command,
)
from cli.commands.knowledge.synthesis.list.command import (
    knowledge_synthesis_list_command,
)
from cli.commands.knowledge.synthesis.refresh.command import (
    knowledge_synthesis_refresh_command,
)
from cli.commands.tool_configs.create.command import tool_configs_create_command
from cli.commands.tool_configs.delete.command import tool_configs_delete_command
from cli.commands.tool_configs.get.command import tool_configs_get_command
from cli.commands.tool_configs.list.command import tool_configs_list_command
from cli.commands.tool_configs.update.command import tool_configs_update_command
from cli.commands.pipelines.list.command import pipelines_list_command
from cli.commands.pipelines.get.command import pipelines_get_command
from cli.commands.pipelines.execute.command import pipelines_execute_command
from cli.commands.pipelines.executions.list_command import (
    pipelines_executions_list_command,
)
from cli.commands.pipelines.executions.get_command import (
    pipelines_executions_get_command,
)
from cli.commands.pipelines.schedules.list_command import (
    pipelines_schedules_list_command,
)
from cli.commands.pipelines.schedules.get_command import pipelines_schedules_get_command
from cli.commands.pipelines.schedules.delete_command import (
    pipelines_schedules_delete_command,
)
from cli.commands.pipelines.costs.command import pipelines_costs_command
from cli.commands.pipelines.approvals.get_command import (
    pipelines_approvals_get_command,
)
from cli.commands.pipelines.approvals.decide_command import (
    pipelines_approvals_decide_command,
)
from cli.commands.pipelines.executions.rerun_command import (
    pipelines_executions_rerun_command,
)
from cli.commands.pipelines.costs.daily_command import pipeline_costs_daily_command
from cli.commands.sessions.list.command import sessions_list_command
from cli.commands.sessions.get.command import sessions_get_command
from cli.commands.sessions.delete.command import sessions_delete_command
from cli.commands.sessions.messages.command import sessions_messages_command
from cli.commands.tags.list.command import tags_list_command
from cli.commands.tags.create.command import tags_create_command
from cli.commands.tags.delete.command import tags_delete_command
from cli.infrastructure.config import load_config, set_global_overrides
from cli.infrastructure.error_display import display_usage_error
from cli.infrastructure.help_json import build_help_json
from cli.infrastructure.help_markdown import render_help_markdown
from cli.infrastructure.output import OutputFormat, OutputService
from alloy_runtime_sdk.logging.config import LogConfig, LogFormat, LogLevel, configure

console = Console()


class AlloyTyperGroup(TyperGroup):
    def main(
        self,
        args: Sequence[str] | None = None,
        prog_name: str | None = None,
        complete_var: str | None = None,
        standalone_mode: bool = True,
        windows_expand_args: bool = True,
        **extra: Any,
    ) -> object:
        try:
            if args is None:
                resolved_args = sys.argv[1:]
            else:
                resolved_args = list(args)

            if prog_name is None:
                prog_name = self.name or "alloy"

            self._main_shell_completion(extra, prog_name, complete_var)

            with self.make_context(prog_name, resolved_args, **extra) as ctx:
                rv = self.invoke(ctx)
                if not standalone_mode:
                    return rv
                ctx.exit()
        except EOFError as error:
            click.echo(file=sys.stderr)
            raise click.Abort() from error
        except KeyboardInterrupt as error:
            raise click.exceptions.Exit(130) from error
        except click.ClickException as error:
            command_path = (
                getattr(getattr(error, "ctx", None), "command_path", None)
                or prog_name
                or "alloy"
            )
            display_usage_error(error.format_message(), command_path=command_path)
            if standalone_mode:
                raise SystemExit(error.exit_code) from None
            return error.exit_code
        except OSError as error:
            if error.errno == errno.EPIPE:
                sys.stdout = cast(TextIO, click.utils.PacifyFlushWrapper(sys.stdout))
                sys.stderr = cast(TextIO, click.utils.PacifyFlushWrapper(sys.stderr))
                if standalone_mode:
                    raise SystemExit(1) from None
                return 1
            raise
        except click.exceptions.Exit as error:
            if standalone_mode:
                raise SystemExit(error.exit_code) from None
            return error.exit_code
        except click.Abort:
            if standalone_mode:
                raise SystemExit(1) from None
            raise


# Root CLI application
# Note: no_args_is_help=False and invoke_without_command=True allows
app = typer.Typer(
    name="alloy",
    cls=AlloyTyperGroup,
    help="Alloy Runtime CLI - Unified interface for AI orchestration. Run without arguments to launch TUI. Normal command output supports compact, table, and json formats.",
    no_args_is_help=False,
    invoke_without_command=True,
    add_help_option=False,
)


@app.callback(invoke_without_command=True)
def global_options(
    ctx: typer.Context,
    help_requested: bool = typer.Option(
        False,
        "--help",
        is_eager=True,
        help="Show this message and exit.",
    ),
    help_json: bool = typer.Option(
        False,
        "--help-json",
        is_eager=True,
        help="Show machine-readable CLI help as JSON and exit.",
    ),
    help_markdown: bool = typer.Option(
        False,
        "--help-markdown",
        is_eager=True,
        help="Show machine-readable CLI help as Markdown and exit.",
    ),
    help_toon: bool = typer.Option(
        False,
        "--help-toon",
        is_eager=True,
        help="Show machine-readable CLI help as TOON and exit.",
    ),
    api_url: Optional[str] = typer.Option(
        None,
        "--api-url",
        help="Override API URL",
        envvar="ALLOY_RUNTIME_API_URL",
    ),
    api_key: Optional[str] = typer.Option(
        None,
        "--api-key",
        help="Override API key",
        envvar="ALLOY_RUNTIME_API_KEY",
    ),
    output_format: OutputFormat = typer.Option(
        OutputFormat.COMPACT,
        "--format",
        help="Choose command output format",
    ),
    no_tui: bool = typer.Option(
        False,
        "--no-tui",
        help="Show CLI help instead of TUI",
    ),
) -> None:
    OutputService.configure(output_format)

    api_url_source = ctx.get_parameter_source("api_url")
    api_key_source = ctx.get_parameter_source("api_key")
    set_global_overrides(
        api_url=api_url if api_url_source is ParameterSource.COMMANDLINE else None,
        api_key=api_key if api_key_source is ParameterSource.COMMANDLINE else None,
    )

    help_format = _resolve_help_format()
    structured_help_format = _resolve_structured_help_format(
        help_json=help_json,
        help_markdown=help_markdown,
        help_toon=help_toon,
        env_help_format=help_format,
        help_requested=help_requested,
        no_tui=no_tui,
    )

    if help_requested:
        if structured_help_format is not None:
            _print_structured_help(ctx, structured_help_format)
        else:
            console.print(ctx.get_help(), markup=False, highlight=False)
        raise typer.Exit()

    if help_json:
        _print_structured_help(ctx, "json")
        raise typer.Exit()

    if help_markdown:
        _print_structured_help(ctx, "markdown")
        raise typer.Exit()

    if help_toon:
        _print_structured_help(ctx, "toon")
        raise typer.Exit()

    if ctx.invoked_subcommand is None:
        if structured_help_format is not None:
            _print_structured_help(ctx, structured_help_format)
            raise typer.Exit()

        if no_tui:
            console.print(ctx.get_help(), markup=False, highlight=False)
            raise typer.Exit()

        _launch_tui()
        raise typer.Exit()


def _resolve_help_format() -> str | None:
    value = os.environ.get("ALLOY_RUNTIME_HELP_FORMAT")
    if value is None:
        return None
    normalized = value.strip().lower()
    if not normalized:
        return None
    return normalized


def _resolve_structured_help_format(
    *,
    help_json: bool,
    help_markdown: bool,
    help_toon: bool,
    env_help_format: str | None,
    help_requested: bool,
    no_tui: bool,
) -> str | None:
    if help_json:
        return "json"
    if help_markdown:
        return "markdown"
    if help_toon:
        return "toon"
    if not (help_requested or no_tui):
        return None
    if env_help_format in {"json", "markdown", "toon"}:
        return env_help_format
    return None


def _print_structured_help(ctx: typer.Context, output_format: str) -> None:
    root_ctx = ctx.find_root()
    root_ctx.command = get_command(app)
    payload = build_help_json(root_ctx)
    if output_format == "json":
        typer.echo(json.dumps(payload, indent=2, sort_keys=False))
        return
    if output_format == "toon":
        module = import_module("cli.infrastructure.help_toon")
        render_help_toon = getattr(module, "render_help_toon")
        help_toon_error = getattr(module, "HelpToonUnavailableError")
        install_hint = getattr(module, "TOON_FORMAT_INSTALL_HINT")
        try:
            typer.echo(render_help_toon(payload), nl=False)
        except help_toon_error as error:
            display_usage_error(
                str(error),
                command_path=root_ctx.command_path,
                hint=install_hint,
            )
            raise typer.Exit(code=1)
        return
    typer.echo(render_help_markdown(payload), nl=False)


def _launch_tui() -> None:
    """Launch the unified Terminal UI application."""
    from cli.tui.app import run_app

    try:
        config = load_config()
    except Exception as e:
        console.print(f"Configuration error: {e}", markup=False, highlight=False)
        console.print(
            "\nRun 'alloy login' to authenticate first.",
            markup=False,
            highlight=False,
        )
        raise typer.Exit(code=1)

    if not config.api_key:
        console.print("Not authenticated.", markup=False, highlight=False)
        console.print(
            "\nRun 'alloy login' to authenticate first.",
            markup=False,
            highlight=False,
        )
        raise typer.Exit(code=1)

    run_app(api_url=config.api_url, api_key=config.api_key)


def _normalize_trogon_default_value(value: object) -> object:
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, tuple):
        tuple_values = cast(tuple[object, ...], value)
        normalized_tuple_items: list[object] = []
        for item in tuple_values:
            normalized_tuple_items.append(_normalize_trogon_default_value(item))
        return tuple(normalized_tuple_items)
    if isinstance(value, list):
        list_values = cast(list[object], value)
        normalized_list_items: list[object] = []
        for item in list_values:
            normalized_list_items.append(_normalize_trogon_default_value(item))
        return normalized_list_items
    return value


@contextmanager
def patched_trogon_enum_defaults() -> Iterator[None]:
    from trogon.introspect import MultiValueParamData

    original_process_cli_option = cast(
        Callable[[object], MultiValueParamData],
        cast(Any, MultiValueParamData).process_cli_option,
    )

    def patched_process_cli_option(value: object) -> Any:
        return original_process_cli_option(_normalize_trogon_default_value(value))

    setattr(
        MultiValueParamData,
        "process_cli_option",
        staticmethod(patched_process_cli_option),
    )
    try:
        yield
    finally:
        setattr(
            MultiValueParamData,
            "process_cli_option",
            staticmethod(original_process_cli_option),
        )


def _launch_command_tui() -> None:
    from trogon import typer as trogon_typer

    with patched_trogon_enum_defaults():
        trogon_typer.Trogon(
            typer.main.get_group(app),
            click_context=click.get_current_context(),
        ).run()


# Register top-level auth commands for convenience
app.command("login")(login_command)
app.command("signup")(signup_command)
app.command("ping")(ping_command)

# Credentials command group
credentials_app = typer.Typer(
    help="Manage supported and configured credentials",
    no_args_is_help=True,
)

# Register credentials group with main app
app.add_typer(credentials_app, name="credentials")

# Register credential commands
credentials_app.command("create")(credentials_create_command)
credentials_app.command("list")(credentials_list_command)
credentials_app.command("supported", help="List supported credential targets")(
    credentials_supported_command
)
credentials_app.command("update")(credentials_update_command)

api_keys_app = typer.Typer(
    help="Manage API keys",
    no_args_is_help=True,
)

app.add_typer(api_keys_app, name="api-keys")

api_keys_app.command("create")(api_keys_create_command)

# Models command group
models_app = typer.Typer(
    help="List and query AI models",
    no_args_is_help=True,
)

# Register models group with main app
app.add_typer(models_app, name="models")

# Register model commands
models_app.command("list")(models_list_command)

# ========================================
# Admin group (system administration)
# ========================================
admin_app = typer.Typer(
    name="admin",
    help="System administration commands (requires system admin privileges)",
    no_args_is_help=True,
)

# Register admin group with main app
app.add_typer(admin_app, name="admin")

# Register admin commands
admin_app.command("bootstrap")(bootstrap_admin_command)

# Admin credentials subgroup
admin_credentials_app = typer.Typer(
    help="Manage system credentials",
    no_args_is_help=True,
)

# Register admin credentials subgroup
admin_app.add_typer(admin_credentials_app, name="credentials")

# Register admin credentials commands
admin_credentials_app.command("create")(admin_credentials_create_command)
admin_credentials_app.command("grant")(admin_credentials_grant_command)

# Admin sync subgroup
admin_sync_app = typer.Typer(
    help="Sync data from external sources",
    no_args_is_help=True,
)

# Register admin sync subgroup
admin_app.add_typer(admin_sync_app, name="sync")

# Register admin sync commands
admin_sync_app.command("models")(sync_command)

# Agents command group
agents_app = typer.Typer(
    help="Manage AI agents",
    no_args_is_help=True,
)

# Register agents group with main app
app.add_typer(agents_app, name="agents")

# Register agent commands
agents_app.command("create")(agents_create_command)
agents_app.command("delete")(agents_delete_command)
agents_app.command("list")(agents_list_command)
agents_app.command("update")(agents_update_command)
agents_app.command("get")(agents_get_command)

# Templates command group
templates_app = typer.Typer(
    help="Manage templates (system instructions, user messages, fragments)",
    no_args_is_help=True,
)

# Register templates group with main app
app.add_typer(templates_app, name="templates")

# Register template commands
templates_app.command("create")(templates_create_command)
templates_app.command("delete")(templates_delete_command)
templates_app.command("get")(templates_get_command)
templates_app.command("list")(templates_list_command)
templates_app.command("render")(templates_render_command)
templates_app.command("validate")(templates_validate_command)
templates_app.command("update")(templates_update_command)
templates_app.command("version")(templates_version_command)
templates_app.command("get-by-version")(templates_get_by_version_command)

# Generate command group
generate_app = typer.Typer(
    help="Generate AI content (text, images, etc.)",
    no_args_is_help=True,
)

# Register generate group with main app
app.add_typer(generate_app, name="generate")

# Register generate commands
generate_app.command("text")(generate_text_command)
generate_app.command("status")(generate_status_command)
generate_app.command("cancel")(generate_cancel_command)

# Content command group
content_app = typer.Typer(
    help="Query and manage content from user and assistant messages",
    no_args_is_help=True,
)

# Register content group with main app
app.add_typer(content_app, name="content")

# Register content commands
content_app.command("list")(content_list_command)
content_app.command("get")(content_get_command)
content_app.command("edit")(content_edit_command)
content_app.command("delete")(content_delete_command)

# Schemas command group
schemas_app = typer.Typer(
    help="Manage input/output validation schemas",
    no_args_is_help=True,
)

# Register schemas group with main app
app.add_typer(schemas_app, name="schemas")

# Register schema commands
schemas_app.command("create")(schemas_create_command)
schemas_app.command("get")(schemas_get_command)
schemas_app.command("list")(schemas_list_command)
schemas_app.command("update")(schemas_update_command)
schemas_app.command("delete")(schemas_delete_command)

# Render command group
render_app = typer.Typer(
    help="Render content to various formats (HTML to image, etc.)",
    no_args_is_help=True,
)

# Register render group with main app
app.add_typer(render_app, name="render")

# Register render commands
render_app.command("image")(render_html_to_image_command)

# Knowledge command group
knowledge_app = typer.Typer(
    help="Manage knowledge collections and documents",
    no_args_is_help=True,
)

# Register knowledge group with main app
app.add_typer(knowledge_app, name="knowledge")

# Knowledge collections subgroup
knowledge_collections_app = typer.Typer(
    help="Manage document collections",
    no_args_is_help=True,
)

# Register collections subgroup with knowledge app
knowledge_app.add_typer(knowledge_collections_app, name="collections")

# Register knowledge collection commands
knowledge_collections_app.command("create")(knowledge_collections_create_command)
knowledge_collections_app.command("delete")(knowledge_collections_delete_command)
knowledge_collections_app.command("get")(knowledge_collections_get_command)
knowledge_collections_app.command("list")(knowledge_collections_list_command)
knowledge_collections_app.command("update")(knowledge_collections_update_command)
knowledge_collections_app.command("cluster")(knowledge_collections_cluster_command)
knowledge_collections_app.command("cluster-status")(
    knowledge_collections_cluster_status_command
)
knowledge_collections_app.command("status")(knowledge_collections_status_command)

# Knowledge documents subgroup
knowledge_documents_app = typer.Typer(
    help="Manage documents in collections",
    no_args_is_help=True,
)

# Register documents subgroup with knowledge app
knowledge_app.add_typer(knowledge_documents_app, name="documents")

# Register knowledge document commands
knowledge_documents_app.command("delete")(knowledge_documents_delete_command)
knowledge_documents_app.command("get")(knowledge_documents_get_command)
knowledge_documents_app.command("ingest")(knowledge_documents_ingest_command)
knowledge_documents_app.command("list")(knowledge_documents_list_command)
knowledge_documents_app.command("count")(knowledge_documents_count_command)
knowledge_documents_app.command("update")(knowledge_documents_update_command)
knowledge_documents_app.command("reingest")(knowledge_documents_reingest_command)
knowledge_documents_app.command("reingest-failed")(
    knowledge_documents_bulk_reingest_failed_command
)
knowledge_documents_app.command("reingest-failed-status")(
    knowledge_documents_bulk_reingest_failed_status_command
)
knowledge_documents_app.command("bulk-metadata")(
    knowledge_documents_bulk_metadata_command
)

# Knowledge synthesis subgroup
knowledge_synthesis_app = typer.Typer(
    help="Manage corpus-level syntheses",
    no_args_is_help=True,
)

# Register synthesis subgroup with knowledge app
knowledge_app.add_typer(knowledge_synthesis_app, name="synthesis")

# Register knowledge synthesis commands
knowledge_synthesis_app.command("create")(knowledge_synthesis_create_command)
knowledge_synthesis_app.command("delete")(knowledge_synthesis_delete_command)
knowledge_synthesis_app.command("get")(knowledge_synthesis_get_command)
knowledge_synthesis_app.command("list")(knowledge_synthesis_list_command)
knowledge_synthesis_app.command("refresh")(knowledge_synthesis_refresh_command)
knowledge_synthesis_app.command("update")(knowledge_synthesis_update_command)

# Register top-level knowledge commands
knowledge_app.command("recover")(knowledge_recover_command)
knowledge_app.command("search")(knowledge_search_command)

# Tool configs command group
tool_configs_app = typer.Typer(
    help="Manage reusable tool configurations",
    no_args_is_help=True,
)

# Register tool-configs group with main app
app.add_typer(tool_configs_app, name="tool-configs")

# Register tool-configs commands
tool_configs_app.command("create")(tool_configs_create_command)
tool_configs_app.command("delete")(tool_configs_delete_command)
tool_configs_app.command("get")(tool_configs_get_command)
tool_configs_app.command("list")(tool_configs_list_command)
tool_configs_app.command("update")(tool_configs_update_command)

executions_app = typer.Typer(
    help="Inspect model executions",
    no_args_is_help=True,
)

app.add_typer(executions_app, name="executions")

executions_app.command("get")(executions_get_command)

# Pipelines command group
pipelines_app = typer.Typer(
    help="Manage and execute pipelines",
    no_args_is_help=True,
)

# Register pipelines group with main app
app.add_typer(pipelines_app, name="pipelines")

# Register pipeline commands
pipelines_app.command("list")(pipelines_list_command)
pipelines_app.command("get")(pipelines_get_command)
pipelines_app.command("execute")(pipelines_execute_command)
pipelines_app.command("costs")(pipelines_costs_command)
pipelines_app.command("costs-daily")(pipeline_costs_daily_command)
pipelines_app.command("create")(pipelines_create_command)
pipelines_app.command("update", help="Update an existing pipeline")(
    pipelines_update_command
)
pipelines_app.command("env")(pipelines_env_vars_command)

# Pipelines executions subgroup
pipelines_executions_app = typer.Typer(
    help="Manage pipeline executions",
    no_args_is_help=True,
)

# Register executions subgroup with pipelines app
pipelines_app.add_typer(pipelines_executions_app, name="executions")

# Register pipeline execution commands
pipelines_executions_app.command("list")(pipelines_executions_list_command)
pipelines_executions_app.command("get")(pipelines_executions_get_command)
pipelines_executions_app.command("rerun")(pipelines_executions_rerun_command)
pipelines_executions_app.command("update")(pipeline_execution_update_command)
pipelines_executions_app.command("costs")(pipeline_execution_costs_command)
pipelines_executions_app.command("costs-by-model")(
    pipeline_execution_costs_by_model_command
)

# Pipelines schedules subgroup
pipelines_schedules_app = typer.Typer(
    help="Manage pipeline schedules",
    no_args_is_help=True,
)

# Register schedules subgroup with pipelines app
pipelines_app.add_typer(pipelines_schedules_app, name="schedules")

# Register pipeline schedule commands
pipelines_schedules_app.command("list")(pipelines_schedules_list_command)
pipelines_schedules_app.command("get")(pipelines_schedules_get_command)
pipelines_schedules_app.command("delete")(pipelines_schedules_delete_command)
pipelines_schedules_app.command("env")(pipelines_schedules_env_vars_command)
pipelines_schedules_app.command("create")(pipelines_schedules_create_command)
pipelines_schedules_app.command("update")(pipelines_schedules_update_command)
pipelines_schedules_app.command("once")(pipelines_schedules_once_command)


# Pipelines approvals subgroup
pipelines_approvals_app = typer.Typer(
    help="Manage approval checkpoints for human-in-the-loop pipelines",
    no_args_is_help=True,
)

# Register approvals subgroup with pipelines app
pipelines_app.add_typer(pipelines_approvals_app, name="approvals")

# Register pipeline approval commands
pipelines_approvals_app.command("get")(pipelines_approvals_get_command)
pipelines_approvals_app.command("decide")(pipelines_approvals_decide_command)

# Sessions command group
sessions_app = typer.Typer(
    help="Manage chat sessions",
    no_args_is_help=True,
)

# Register sessions group with main app
app.add_typer(sessions_app, name="sessions")

# Register session commands
sessions_app.command("list")(sessions_list_command)
sessions_app.command("get")(sessions_get_command)
sessions_app.command("delete")(sessions_delete_command)
sessions_app.command("messages")(sessions_messages_command)

# Tags command group
tags_app = typer.Typer(
    help="Manage tags for organizing content",
    no_args_is_help=True,
)

# Register tags group with main app
app.add_typer(tags_app, name="tags")

# Register tag commands
tags_app.command("list")(tags_list_command)
tags_app.command("create")(tags_create_command)
tags_app.command("delete")(tags_delete_command)
tags_app.command("get")(tags_get_command)
tags_app.command("update")(tags_update_command)

# Billing command group
billing_app = typer.Typer(
    help="Manage billing projects and costs",
    no_args_is_help=True,
)

app.add_typer(billing_app, name="billing")

# Billing projects subgroup
billing_projects_app = typer.Typer(
    help="Manage billing projects",
    no_args_is_help=True,
)

billing_app.add_typer(billing_projects_app, name="projects")

billing_projects_app.command("create")(billing_projects_create_command)
billing_projects_app.command("list")(billing_projects_list_command)
billing_projects_app.command("get")(billing_projects_get_command)

# Billing costs subgroup
billing_costs_app = typer.Typer(
    help="View billing costs",
    no_args_is_help=True,
)

billing_app.add_typer(billing_costs_app, name="costs")

billing_costs_app.command("summary")(billing_costs_summary_command)
billing_costs_app.command("daily")(billing_costs_daily_command)
billing_costs_app.command("by-agent")(billing_costs_by_agent_command)
billing_costs_app.command("by-model")(billing_costs_by_model_command)

# Tools command group
tools_app = typer.Typer(
    help="List and inspect available tools",
    no_args_is_help=True,
)

app.add_typer(tools_app, name="tools")

tools_app.command("list")(tools_list_command)
tools_app.command("get")(tools_get_command)

# Audio command group
audio_app = typer.Typer(
    help="Audio processing commands",
    no_args_is_help=True,
)

app.add_typer(audio_app, name="audio")

audio_app.command("transcribe")(audio_transcribe_command)

# Organizations command group
organizations_app = typer.Typer(
    help="Manage organizations",
    no_args_is_help=True,
)

app.add_typer(organizations_app, name="organizations")

organizations_app.command("create")(organizations_create_command)

# Users command group
users_app = typer.Typer(
    help="Manage users",
    no_args_is_help=True,
)

app.add_typer(users_app, name="users")

users_app.command("create")(users_create_command)


app.command("tui", help="Open Textual TUI.")(_launch_command_tui)


def main() -> None:
    """CLI entry point."""
    # Configure logging for CLI (logs to file, Rich console for user output)
    log_level_str = os.environ.get("LOG_LEVEL", "warning").lower()
    log_level = LogLevel(log_level_str)

    # Determine log file path - use root logs directory
    # From cli/main.py -> cli -> packages -> core-service
    project_root = Path(__file__).parent.parent.parent.parent
    log_dir = project_root / "logs"
    log_file = log_dir / "cli.log"

    configure(
        LogConfig(
            level=log_level,
            format=LogFormat.JSON,  # Always JSON for parsability
            service_name="alloy-runtime-cli",
            output_file=str(log_file),
            capture_stdlib_logging=True,
        )
    )

    app()


if __name__ == "__main__":
    main()



"""Presenter for tool config update command output."""

from cli.commands.tool_configs.get.presenter import present_tool_config_details
from cli.infrastructure.output import OutputService
from alloy_runtime_types.dtos.tool_configs import UpdateToolConfigResponse


def present_tool_config_update_success(response: UpdateToolConfigResponse) -> None:
    """Present successful tool config update.

    Args:
        response: Tool config update response from server
    """
    output = OutputService.get()

    output.result(
        action="update",
        resource="tool_config",
        identifier=str(response.id),
        name=response.name,
    )
    present_tool_config_details(response)



"""Plain text renderer for chat messages.

Renders messages without any Rich markup formatting.
"""

from datetime import datetime

from alloy_runtime_types.dtos.sessions import MessageResponse

from cli.tui.chat.renderers.base import get_message_body
from cli.tui.chat.types import format_time


class PlainTextRenderer:
    """Renders messages as plain text without formatting.

    Useful for:
        - Exporting conversations
        - Copying to clipboard
        - Accessibility contexts
        - Testing and debugging

    """

    def render(
        self,
        messages: tuple[MessageResponse, ...],
        pending_user_message: str | None = None,
        assistant_name: str = "Assistant",
        is_generating: bool = False,
        thinking_collapsed: bool = False,
    ) -> str:
        """Render messages to a plain text string.

        Args:
            messages: Tuple of message responses to render.
            pending_user_message: User message shown optimistically before server confirms.
            assistant_name: Name to display for assistant (agent name or model).
            is_generating: Whether currently waiting for the final response.
            thinking_collapsed: Whether thinking blocks are collapsed (ignored in plain).

        Returns:
            Plain text string without Rich markup.
        """
        lines: list[str] = []

        for msg in messages:
            time_str = format_time(msg.created_at)

            if msg.role == "user":
                body = get_message_body(msg)
                lines.append(f"You [{time_str}]:")
            else:
                body = get_message_body(msg)
                lines.append(f"{assistant_name} [{time_str}]:")

            lines.append(body)
            lines.append("")

        # Show pending user message
        if pending_user_message:
            now_str = datetime.now().strftime("%H:%M")
            lines.append(f"You [{now_str}]:")
            lines.append(pending_user_message)
            lines.append("")

        # Show in-flight response placeholder
        if is_generating:
            lines.append(f"{assistant_name}:")
            lines.append("Waiting for response...")

        return "\n".join(lines)



"""Knowledge synthesis delete command implementation."""

import typer

from cli.commands.flag_utils import validate_uuid
from cli.commands.knowledge.synthesis.delete.presenter import (
    present_synthesis_delete_success,
)
from cli.infrastructure.command import async_command, authenticated_client


def knowledge_synthesis_delete_command(
    synthesis_id: str = typer.Argument(
        ...,
        help="Synthesis UUID to delete",
    ),
    yes: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip confirmation prompt",
    ),
) -> None:
    """Delete a synthesis.

    This permanently removes the synthesis, its collection associations,
    and its source tracking records.

    Examples:
        alloy knowledge synthesis delete <synthesis-uuid>
        alloy knowledge synthesis delete <uuid> --yes
    """
    # Validate synthesis UUID
    validate_uuid(synthesis_id, "synthesis")

    # Confirm deletion unless --yes flag provided
    if not yes:
        confirm = typer.confirm(
            f"Are you sure you want to delete synthesis {synthesis_id}?"
        )
        if not confirm:
            raise typer.Abort()

    _execute_delete(synthesis_id)


@async_command
async def _execute_delete(synthesis_id: str) -> None:
    """Execute synthesis deletion."""
    async with authenticated_client() as (_config, client):
        response = await client.delete_synthesis(synthesis_id)

    present_synthesis_delete_success(response)



"""Pydantic models for TUI screen data types.

This module contains view models used by screens to represent
data in a structured, type-safe manner.
"""

from datetime import datetime
from typing import Any, Literal, cast

from pydantic import BaseModel, Field


def _tag_list_factory() -> list["TagView"]:
    return []


class TagView(BaseModel):
    """View model for a tag."""

    tag_path: str


class ExecutionMetadataView(BaseModel):
    """View model for execution metadata."""

    external_id: str | None = None
    provider_key: str = ""
    provider_model_name: str = ""


class ContentPartView(BaseModel):
    """View model for content parts in the Content browser.

    This provides a flattened, UI-friendly representation of
    ContentPartItemResponse for use in the ContentScreen.
    """

    id: str = Field(description="Content part UUID as string")
    message_role: Literal["user", "assistant"] = Field(
        description="Parent message role for the content part"
    )
    content_type_id: str = Field(description="Content type (text, code, json, etc.)")
    content_text: str | None = Field(
        default=None, description="Plain text content (null for structured/media types)"
    )
    content_structured: dict[str, Any] | None = Field(
        default=None, description="Structured JSON content (null for text types)"
    )
    tags: list[TagView] = Field(
        default_factory=_tag_list_factory,
        description="Associated tags",
    )
    execution_metadata: ExecutionMetadataView | None = Field(
        default=None, description="Execution metadata"
    )
    created_at: datetime | None = Field(default=None, description="Creation timestamp")
    updated_at: datetime | None = Field(
        default=None, description="Last update timestamp"
    )

    @property
    def external_id(self) -> str | None:
        """Get the external_id from execution metadata if available."""
        if self.execution_metadata:
            return self.execution_metadata.external_id
        return None

    @classmethod
    def from_response(cls, part: Any) -> "ContentPartView":
        """Create a ContentPartView from a ContentPartItemResponse.

        Args:
            part: A ContentPartItemResponse from the server client

        Returns:
            A ContentPartView instance
        """
        raw_tags = cast(list[Any], part.tags or [])
        tags: list[TagView] = []
        for tag in raw_tags:
            tags.append(TagView(tag_path=cast(str, tag.tag_path)))
        exec_meta = None
        if part.execution_metadata:
            exec_meta = ExecutionMetadataView(
                external_id=part.execution_metadata.external_id,
                provider_key=part.execution_metadata.provider_key,
                provider_model_name=part.execution_metadata.provider_model_name,
            )

        return cls(
            id=str(part.id),
            message_role=part.message_role,
            content_type_id=part.content_type_id,
            content_text=part.content_text,
            content_structured=part.content_structured,
            tags=tags,
            execution_metadata=exec_meta,
            created_at=part.created_at,
            updated_at=part.updated_at,
        )



"""Presenter for agent creation output."""

from cli.commands.agents.create.fields import (
    format_generation_params,
    format_models,
    format_tags,
    format_tools,
)
from cli.infrastructure.formatting.fields import (
    format_datetime,
    format_optional,
    format_uuid,
)
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig
from alloy_runtime_types.dtos.agents import CreateAgentResponse


def present_agent_create_success(response: CreateAgentResponse) -> None:
    """Present successful agent creation.

    Args:
        response: Agent creation response from server
    """
    output = OutputService.get()
    output.result(
        action="create",
        resource="agent",
        identifier=str(response.id),
        name=response.name,
    )
    system_prompt_id = (
        response.system_instruction_template_id
        or response.system_instruction_version_id
    )

    fields = [
        FieldConfig("id", "ID", lambda v: format_uuid(v, short=False)),
        FieldConfig("name", "Name"),
        FieldConfig("description", "Description", format_optional),
        FieldConfig("models", "Models", lambda _: format_models(response)),
        FieldConfig("tools", "Tools", lambda _: format_tools(response)),
        FieldConfig("tags", "Tags", lambda _: format_tags(response)),
        FieldConfig(
            "generation_params",
            "Generation Params",
            lambda _: format_generation_params(response),
        ),
        FieldConfig(
            "system_instruction_template_id",
            "System Prompt ID",
            lambda _: format_optional(
                system_prompt_id, lambda x: format_uuid(x, short=False)
            ),
        ),
        FieldConfig(
            "input_schema_id",
            "Input Schema ID",
            lambda v: format_optional(v, lambda x: format_uuid(x, short=False)),
        ),
        FieldConfig(
            "output_schema_id",
            "Output Schema ID",
            lambda v: format_optional(v, lambda x: format_uuid(x, short=False)),
        ),
        FieldConfig(
            "organization_id",
            "Organization ID",
            lambda v: format_optional(v, lambda x: format_uuid(x, short=True)),
        ),
        FieldConfig(
            "user_id",
            "User ID",
            lambda v: format_optional(v, lambda x: format_uuid(x, short=True)),
        ),
        FieldConfig("created_at", "Created At", format_datetime),
    ]

    output.entity(response, f"Agent: {response.name}", fields, raw_payload=response)



"""Pipeline schedule get command implementation."""

from uuid import UUID

import typer

from cli.commands.flag_utils import validate_uuid
from cli.commands.pipelines.schedules.presenter import present_schedule_detail
from cli.infrastructure.command import async_command, authenticated_client


def pipelines_schedules_get_command(
    pipeline_id: str = typer.Argument(
        ...,
        help="Pipeline UUID",
    ),
    schedule_id: str = typer.Argument(
        ...,
        help="Schedule UUID",
    ),
) -> None:
    """Get details of a pipeline schedule.

    Examples:
        alloy pipelines schedules get 123e4567... 456e7890...
    """
    p_uuid = validate_uuid(pipeline_id, "pipeline")
    s_uuid = validate_uuid(schedule_id, "schedule")
    _execute_get(pipeline_id=p_uuid, schedule_id=s_uuid)


@async_command
async def _execute_get(pipeline_id: UUID, schedule_id: UUID) -> None:
    """Execute get schedule operation."""
    async with authenticated_client() as (_config, client):
        response = await client.get_pipeline_schedule(pipeline_id, schedule_id)

    present_schedule_detail(response)



"""Tags update command implementation."""

from typing import Optional
from uuid import UUID

import typer

from cli.commands.tags.update.presenter import present_update_result
from cli.commands.flag_utils import validate_uuid, require_update_flags
from cli.infrastructure.command import async_command, authenticated_client
from alloy_runtime_types.dtos.tags import UpdateTagRequest


def tags_update_command(
    tag_id: str = typer.Argument(
        ...,
        help="Tag UUID",
    ),
    display_name: Optional[str] = typer.Option(
        None,
        "-n",
        "--display-name",
        help="New display name",
    ),
    description: Optional[str] = typer.Option(
        None,
        "-d",
        "--description",
        help="New description",
    ),
) -> None:
    """Update a tag's display name or description.

    Examples:
        alloy tags update 123e4567-89ab... -n "New Name"
        alloy tags update 123e4567-89ab... -d "New description"
    """
    require_update_flags(
        ("--display-name", display_name),
        ("--description", description),
    )

    tag_uuid = validate_uuid(tag_id, "tag")
    _execute_update(
        tag_id=tag_uuid,
        display_name=display_name,
        description=description,
    )


@async_command
async def _execute_update(
    tag_id: UUID,
    display_name: Optional[str],
    description: Optional[str],
) -> None:
    request = UpdateTagRequest(
        display_name=display_name,
        description=description,
    )
    async with authenticated_client() as (_config, client):
        response = await client.update_tag(tag_id, request)

    present_update_result(response)



"""Credential update command implementation."""

from typing import Optional
from uuid import UUID

import typer

from cli.commands.credentials.update.presenter import present_credential_update_success
from cli.commands.shared_flags import (
    CREDENTIAL_VALUE_OPTIONAL,
    DESCRIPTION,
    NAME_OPTIONAL,
)
from cli.infrastructure.command import async_command, authenticated_client
from alloy_runtime_types.dtos.credentials import UpdateCredentialRequest


def credentials_update_command(
    credential_id: UUID = typer.Argument(..., help="Credential UUID"),
    name: Optional[str] = NAME_OPTIONAL,
    description: Optional[str] = DESCRIPTION,
    value: Optional[str] = CREDENTIAL_VALUE_OPTIONAL,
) -> None:
    """Update a credential (system or organization).

    At least one flag required.

    Examples:

        alloy credentials update <uuid> -n "Production Key"
        alloy credentials update <uuid> -d "Updated" --value "sk_new"
    """
    _execute_update(credential_id, name, description, value)


@async_command
async def _execute_update(
    credential_id: UUID,
    name: Optional[str],
    description: Optional[str],
    value: Optional[str],
) -> None:
    """Execute the credential update operation."""
    request = UpdateCredentialRequest(
        name=name,
        description=description,
        secret=value,
    )

    async with authenticated_client() as (_config, client):
        response = await client.update_credential(credential_id, request)

    present_credential_update_success(response)



"""Streaming text renderer with thinking block support."""

import sys
from collections.abc import AsyncGenerator

from pydantic import BaseModel
from rich.console import Console

from alloy_runtime_types.dtos.generation import GenerateTextStreamChunk
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig


class StreamRenderResult(BaseModel):
    content: str
    thinking: str | None = None
    session_id: str | None = None


class StreamRenderer:
    """Renders streaming text generation with thinking block support."""

    def __init__(self, console: Console, stderr_console: Console | None = None):
        """Initialize stream renderer.

        Args:
            console: Rich console for output
        """
        self.console = console
        self.stderr_console = stderr_console or Console(file=sys.stderr)
        self._thinking_text = ""
        self._content_text = ""
        self._session_id: str | None = None
        self._in_thinking_block = False

    async def render_stream(
        self,
        stream: AsyncGenerator[GenerateTextStreamChunk, None],
        show_thinking: bool = True,
    ) -> tuple[str, str, str | None]:
        """Render a streaming generation.

        Args:
            stream: Async generator of stream chunks
            show_thinking: Whether to display thinking blocks (default True)

        Returns:
            Tuple of (content_text, thinking_text, session_id) - accumulated outputs
        """
        self._thinking_text = ""
        self._content_text = ""
        self._session_id = None
        self._in_thinking_block = False

        async for chunk in stream:
            self._process_chunk(chunk, show_thinking)

        # Close thinking block if still open at end of stream
        if self._in_thinking_block:
            self._close_thinking_block()

        return self._content_text, self._thinking_text, self._session_id

    def present_final_output(
        self, content: str, thinking: str, session_id: str | None, show_thinking: bool
    ) -> None:
        result = StreamRenderResult(
            content=content,
            thinking=thinking or None if show_thinking else None,
            session_id=session_id,
        )
        fields = [FieldConfig("content", "Output")]
        if show_thinking and thinking:
            fields.append(FieldConfig("thinking", "Thinking"))
        if session_id:
            fields.append(FieldConfig("session_id", "Session ID"))
        OutputService.get().entity(result, "Generated Text", fields, raw_payload=result)

    def _process_chunk(
        self, chunk: GenerateTextStreamChunk, show_thinking: bool
    ) -> None:
        """Process a stream chunk.

        Args:
            chunk: Stream chunk to process
            show_thinking: Whether to display thinking text
        """
        if chunk.chunk_type == "thinking":
            if chunk.reasoning_content:
                self._thinking_text += chunk.reasoning_content
                if show_thinking:
                    if not self._in_thinking_block:
                        self._open_thinking_block()
                    self.stderr_console.out(
                        chunk.reasoning_content, style="red", end=""
                    )

        elif chunk.chunk_type == "metadata":
            if chunk.metadata and "chat_session_id" in chunk.metadata:
                self._session_id = chunk.metadata["chat_session_id"]

        else:
            if self._in_thinking_block:
                self._close_thinking_block()

            if chunk.content:
                self._content_text += chunk.content

    def _open_thinking_block(self) -> None:
        """Print opening thinking tag to stderr."""
        self.stderr_console.out("\n<thinking>\n", style="red", end="")
        self._in_thinking_block = True

    def _close_thinking_block(self) -> None:
        """Print closing thinking tag to stderr."""
        if not self._thinking_text.endswith("\n"):
            self.stderr_console.out("\n", end="")
        self.stderr_console.out("</thinking>\n\n", style="red", end="")
        self._in_thinking_block = False



"""Command for reloading the current session's messages."""

from dataclasses import dataclass
from uuid import UUID

from alloy_runtime_sdk.api_client.client import ApiClient
from alloy_runtime_types.dtos.sessions import MessageResponse

from cli.tui.chat.commands.base import CommandResult
from cli.tui.chat.store import ChatStore


@dataclass
class ReloadSessionResult:
    """Result data for reload session operation.

    Attributes:
        messages: Updated list of messages.
    """

    messages: list[MessageResponse]


@dataclass
class ReloadSessionCommand:
    """Command to reload the current session's messages.

    This is a lightweight operation that only fetches messages,
    without changing the session context. Useful after sending
    a message or regenerating to get the updated message list.

    Attributes:
        client: Server client for API calls.
        store: Chat store for state management.
        session_id: ID of the session to reload.
        message_limit: Maximum number of messages to fetch.
    """

    client: ApiClient
    store: ChatStore
    session_id: UUID
    message_limit: int = 100

    async def execute(self) -> CommandResult[ReloadSessionResult]:
        """Execute the reload session operation.

        Returns:
            CommandResult with updated messages on success.
        """
        try:
            session = await self.client.get_session(
                self.session_id, limit=self.message_limit
            )
            self.store.set_messages(list(session.messages))

            return CommandResult[ReloadSessionResult](
                success=True,
                data=ReloadSessionResult(messages=list(session.messages)),
            )

        except Exception as e:
            return CommandResult[ReloadSessionResult](success=False, error=str(e))



"""Presenter for tags get command output."""

from typing import Any, cast

from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig
from alloy_runtime_types.dtos.tags import GetTagResponse


def _format_optional_text(value: str | None) -> Any:
    if value is None:
        return cast(str, None)
    return value


def present_tag_details(response: GetTagResponse) -> None:
    output = OutputService.get()
    output.entity(
        response.tag,
        f"Tag: {response.tag.tag_path}",
        [
            FieldConfig("id", "ID", str),
            FieldConfig("tag_path", "Tag Path"),
            FieldConfig("display_name", "Display Name", _format_optional_text),
            FieldConfig("description", "Description", _format_optional_text),
        ],
        raw_payload=response,
    )



"""Configuration management for CLI.

Loads configuration with hierarchy: CLI flags > Environment variables > Config file
"""

import os

from pydantic import BaseModel, Field, field_validator

from cli.infrastructure.auth_storage import load_credentials


# Global storage for CLI flag overrides (set by main.py callback)
_global_api_url: str | None = None
_global_api_key: str | None = None


def set_global_overrides(
    api_url: str | None = None, api_key: str | None = None
) -> None:
    """Set global CLI flag overrides.

    Called by the main app callback to store --api-url and --api-key flags.

    Args:
        api_url: Global --api-url flag value
        api_key: Global --api-key flag value
    """
    global _global_api_url, _global_api_key
    _global_api_url = api_url
    _global_api_key = api_key


def get_api_url(default: str = "http://localhost:8000") -> str:
    """Get API URL using the configuration hierarchy.

    Precedence order (highest to lowest):
    1. Global CLI flag (--api-url)
    2. Environment variable (ALLOY_RUNTIME_API_URL)
    3. Config file (~/.alloy-runtime/config)
    4. Default value

    Args:
        default: Fallback value if no configuration source provides URL

    Returns:
        API URL string

    Example:
        api_url = get_api_url()  # Uses hierarchy with localhost default
    """
    # Try global flag (highest priority)
    if _global_api_url:
        return _global_api_url

    # Try environment variable
    env_url = os.environ.get("ALLOY_RUNTIME_API_URL")
    if env_url:
        return env_url

    # Try config file
    file_config = load_credentials()
    if file_config:
        return file_config[0]

    # Use default
    return default


class CLIConfig(BaseModel):
    """CLI configuration with hierarchical loading.

    Precedence order (highest to lowest):
    1. CLI flags (--api-key, --api-url)
    2. Environment variables (ALLOY_RUNTIME_API_KEY, ALLOY_RUNTIME_API_URL)
    3. Config file (~/.alloy-runtime/config)
    """

    api_url: str = Field(..., description="Alloy Runtime server base URL")
    api_key: str = Field(..., description="API authentication key")
    timeout: float | None = Field(
        default=None, description="Request timeout in seconds"
    )

    @field_validator("api_url")
    @classmethod
    def validate_api_url(cls, v: str) -> str:
        """Ensure API URL is non-empty and properly formatted."""
        _ = cls  # Required by Pydantic validator
        if not v or not v.strip():
            raise ValueError("API URL cannot be empty")
        return v.rstrip("/")  # Remove trailing slash

    @field_validator("api_key")
    @classmethod
    def validate_api_key(cls, v: str) -> str:
        """Ensure API key is non-empty."""
        _ = cls  # Required by Pydantic validator
        if not v or not v.strip():
            raise ValueError("API key cannot be empty")
        return v


class ConfigurationError(Exception):
    """Raised when configuration cannot be loaded."""

    pass


def load_config() -> CLIConfig:
    """Load CLI configuration with hierarchical precedence.

    Precedence order (highest to lowest):
    1. Global CLI flags (--api-key, --api-url set via callback)
    2. Environment variables (ALLOY_RUNTIME_API_KEY, ALLOY_RUNTIME_API_URL)
    3. Config file (~/.alloy-runtime/config)

    Returns:
        Validated CLI configuration

    Raises:
        ConfigurationError: If no configuration source provides required values

    Example:
        # Load with hierarchy
        config = load_config()

        # CLI usage:
        alloy --api-url https://prod.example.com admin sync models
        alloy --api-key sk_prod_... credentials update <id>
    """
    # Load from config file (lowest priority)
    file_config = load_credentials()
    api_url_from_file = file_config[0] if file_config else None
    api_key_from_file = file_config[1] if file_config else None

    # Load from environment variables (middle priority)
    api_url_from_env = os.environ.get("ALLOY_RUNTIME_API_URL")
    api_key_from_env = os.environ.get("ALLOY_RUNTIME_API_KEY")

    # Apply hierarchy: Global CLI flags > env vars > config file
    # Note: Must check "is not None" because flags can be explicitly None when not passed
    if _global_api_url is not None:
        api_url = _global_api_url
    elif api_url_from_env:
        api_url = api_url_from_env
    else:
        api_url = api_url_from_file

    if _global_api_key is not None:
        api_key = _global_api_key
    elif api_key_from_env:
        api_key = api_key_from_env
    else:
        api_key = api_key_from_file

    # Validate that we have the required configuration
    if not api_url:
        raise ConfigurationError(
            "API URL not configured. Please provide via:\n"
            "  1. CLI flag: --api-url http://localhost:8000\n"
            "  2. Environment variable: ALLOY_RUNTIME_API_URL\n"
            "  3. Login first: alloy login"
        )

    if not api_key:
        raise ConfigurationError(
            "API key not configured. Please provide via:\n"
            "  1. CLI flag: --api-key sk_live_...\n"
            "  2. Environment variable: ALLOY_RUNTIME_API_KEY\n"
            "  3. Login first: alloy login"
        )

    # Load optional timeout
    timeout = None
    timeout_str = os.environ.get("ALLOY_RUNTIME_TIMEOUT")
    if timeout_str:
        try:
            timeout = float(timeout_str)
        except ValueError:
            raise ConfigurationError(
                "ALLOY_RUNTIME_TIMEOUT must be a valid number (seconds)"
            )

    try:
        return CLIConfig(api_url=api_url, api_key=api_key, timeout=timeout)
    except ValueError as e:
        raise ConfigurationError(f"Invalid configuration: {e}")



"""Injection service for TUI chat.

Wraps the shared InjectionResolver and provides caching for autocomplete.
This service is designed for use with Textual's async workers.

Supports the same macro syntax as templates:
- @fragment(name)       - Include a fragment template
- @text(name)           - Insert text from a content part
- @json(name)           - Insert JSON from a content part
- @schema(name)         - Reference a schema definition
"""

import logging
import time
from dataclasses import dataclass

from alloy_runtime_sdk.api_client.client import ApiClient

from cli.infrastructure.injection.resolver import (
    InjectionResolver,
    ResolvedMessage,
)

logger = logging.getLogger(__name__)

# Cache TTL in seconds (5 minutes)
CACHE_TTL = 300


@dataclass
class FragmentInfo:
    """Fragment template information for autocomplete."""

    id: str
    name: str
    description: str | None
    content_type_id: str


@dataclass
class ContentPartInfo:
    """Content part information for autocomplete."""

    id: str
    content_type_id: str
    content_text: str | None
    has_structured: bool


@dataclass
class SchemaInfo:
    """Schema information for autocomplete."""

    id: str
    schema_name: str
    version: str
    schema_format_id: str
    description: str | None


@dataclass
class InjectionCompletion:
    """A single injection autocomplete suggestion."""

    text: str  # Full insertion text: @fragment(my-fragment)
    display_name: str  # Display name: "my-fragment"
    description: str  # Description/type info
    injection_type: str  # "fragment", "text", "json", or "schema"


class InjectionService:
    """Manages injection resolution and caching for TUI.

    Features:
    - Wraps InjectionResolver for API calls
    - Caches fragments/content/schemas for autocomplete
    - Provides async methods suitable for Textual workers
    - Shares ApiClient with the rest of the TUI

    Usage:
        service = InjectionService(client)
        await service.refresh_cache()

        # Check for completions
        completions = service.get_completions("fragment", "my-fra")

        # Resolve message before sending
        if has_injections(message):
            result = await service.resolve_message(message)
            if result.has_errors:
                # Handle errors
                ...
            message = result.text
    """

    def __init__(self, client: ApiClient):
        """Initialize the injection service.

        Args:
            client: Shared ApiClient instance from the TUI app.
        """
        self._client = client
        self._resolver = InjectionResolver(client=client)

        # Autocomplete caches
        self._fragment_cache: list[FragmentInfo] = []
        self._content_cache: list[ContentPartInfo] = []
        self._schema_cache: list[SchemaInfo] = []
        self._cache_expiry: float = 0

    @property
    def is_cache_valid(self) -> bool:
        """Check if the autocomplete cache is still valid."""
        return time.time() < self._cache_expiry

    async def resolve_message(self, message: str) -> ResolvedMessage:
        """Resolve all injection patterns in a message.

        Args:
            message: User message with potential injection patterns.

        Returns:
            ResolvedMessage with resolved text and metadata.
        """
        return await self._resolver.resolve(message)

    async def refresh_cache(self) -> None:
        """Refresh all autocomplete caches from the API.

        This should be called:
        - On session load/create
        - Periodically in the background
        - When cache expires and user triggers autocomplete
        """
        try:
            # Fetch all data in parallel
            import asyncio

            results = await asyncio.gather(
                self._fetch_fragments(),
                self._fetch_content(),
                self._fetch_schemas(),
                return_exceptions=True,
            )

            # Log any errors but don't fail
            for i, result in enumerate(results):
                if isinstance(result, Exception):
                    logger.debug(
                        "injection_cache_refresh_partial_failure",
                        extra={"index": i, "error": str(result)},
                    )

            # Update expiry
            self._cache_expiry = time.time() + CACHE_TTL

            logger.debug(
                "injection_cache_refreshed",
                extra={
                    "fragments": len(self._fragment_cache),
                    "content": len(self._content_cache),
                    "schemas": len(self._schema_cache),
                },
            )

        except Exception as e:
            logger.debug(
                "injection_cache_refresh_failed",
                extra={"error": str(e)},
            )

    async def _fetch_fragments(self) -> None:
        """Fetch fragment templates for autocomplete cache.

        Only fetches templates with content_type='fragment'.
        """
        from alloy_runtime_types.enums.template_enums import TemplateContentType

        try:
            response = await self._client.list_templates(
                limit=100, content_type=TemplateContentType.FRAGMENT
            )
            self._fragment_cache = [
                FragmentInfo(
                    id=str(t.id),
                    name=t.name,
                    description=t.description,
                    content_type_id=t.content_type_id,
                )
                for t in response.templates
            ]
        except Exception as e:
            logger.debug("Failed to fetch fragments: %s", e)
            raise

    async def _fetch_content(self) -> None:
        """Fetch content parts for autocomplete cache."""
        try:
            response = await self._client.list_content_parts(limit=100)
            self._content_cache = [
                ContentPartInfo(
                    id=str(p.id),
                    content_type_id=p.content_type_id,
                    content_text=p.content_text,
                    has_structured=p.content_structured is not None,
                )
                for p in response.content_parts
            ]
        except Exception as e:
            logger.debug("Failed to fetch content parts: %s", e)
            raise

    async def _fetch_schemas(self) -> None:
        """Fetch schemas for autocomplete cache."""
        try:
            response = await self._client.list_schemas(limit=100)
            self._schema_cache = [
                SchemaInfo(
                    id=str(s.id),
                    schema_name=s.schema_name,
                    version=str(s.version),
                    schema_format_id=s.schema_format_id,
                    description=s.description,
                )
                for s in response.schemas
            ]
        except Exception as e:
            logger.debug("Failed to fetch schemas: %s", e)
            raise

    def get_completions(
        self,
        injection_type: str,
        partial_identifier: str = "",
        limit: int = 10,
    ) -> list[InjectionCompletion]:
        """Get autocomplete suggestions for a partial injection.

        Args:
            injection_type: Type of injection ("fragment", "text", "json", "schema").
            partial_identifier: Partial identifier typed by user.
            limit: Maximum number of suggestions to return.

        Returns:
            List of InjectionCompletion objects.
        """
        partial_lower = partial_identifier.lower()

        if injection_type == "fragment":
            return self._get_fragment_completions(partial_lower, limit)
        elif injection_type in ("text", "json"):
            return self._get_content_completions(
                partial_lower, injection_type == "json", limit
            )
        elif injection_type == "schema":
            return self._get_schema_completions(partial_lower, limit)
        else:
            return []

    def _get_fragment_completions(
        self, partial: str, limit: int
    ) -> list[InjectionCompletion]:
        """Get fragment completions matching the partial identifier."""
        completions: list[InjectionCompletion] = []

        for fragment in self._fragment_cache:
            name_lower = fragment.name.lower()
            desc_lower = (fragment.description or "").lower()

            if partial in name_lower or partial in desc_lower:
                # Build the insertion text (no quotes, no variables)
                text = f"@fragment({fragment.name})"

                # Format description
                description = fragment.description or "no description"

                completions.append(
                    InjectionCompletion(
                        text=text,
                        display_name=fragment.name,
                        description=description,
                        injection_type="fragment",
                    )
                )

                if len(completions) >= limit:
                    break

        return completions

    def _get_content_completions(
        self, partial: str, is_json: bool, limit: int
    ) -> list[InjectionCompletion]:
        """Get content completions matching the partial identifier."""
        completions: list[InjectionCompletion] = []

        for content in self._content_cache:
            # Match on UUID or content text preview
            id_lower = content.id.lower()
            text_lower = (content.content_text or "").lower()

            if partial in id_lower or partial in text_lower:
                # Build the insertion text (no quotes)
                if is_json:
                    text = f"@json({content.id})"
                    format_label = "JSON"
                else:
                    text = f"@text({content.id})"
                    format_label = "text"

                # Create preview from content text
                if content.content_text:
                    preview = content.content_text.replace("\n", " ").strip()
                else:
                    preview = "{structured}" if content.has_structured else "{empty}"

                description = f"{content.content_type_id} | {format_label}"

                completions.append(
                    InjectionCompletion(
                        text=text,
                        display_name=preview,
                        description=description,
                        injection_type="json" if is_json else "text",
                    )
                )

                if len(completions) >= limit:
                    break

        return completions

    def _get_schema_completions(
        self, partial: str, limit: int
    ) -> list[InjectionCompletion]:
        """Get schema completions matching the partial identifier."""
        completions: list[InjectionCompletion] = []

        for schema in self._schema_cache:
            name_lower = schema.schema_name.lower()
            desc_lower = (schema.description or "").lower()

            if partial in name_lower or partial in desc_lower:
                # Build insertion text (no quotes)
                text = f"@schema({schema.schema_name})"

                # Format description
                desc_preview = schema.description or "no description"
                description = f"{schema.schema_format_id} | {desc_preview}"

                completions.append(
                    InjectionCompletion(
                        text=text,
                        display_name=schema.schema_name,
                        description=description,
                        injection_type="schema",
                    )
                )

                if len(completions) >= limit:
                    break

        return completions

    @property
    def fragments(self) -> list[FragmentInfo]:
        """Get cached fragments."""
        return self._fragment_cache

    @property
    def content_parts(self) -> list[ContentPartInfo]:
        """Get cached content parts."""
        return self._content_cache

    @property
    def schemas(self) -> list[SchemaInfo]:
        """Get cached schemas."""
        return self._schema_cache



"""Copy-on-select widgets for TUI applications.

This module provides widgets that automatically copy selected text to the clipboard.
These are useful for allowing users to select and copy text from the TUI without
needing to use explicit keybindings.

Available widgets:
- CopyOnSelectTextArea: TextArea that copies selected text automatically
- SelectableTextArea: Read-only TextArea for displaying selectable content
"""

import asyncio
import os
import subprocess
from typing import Any, Protocol, cast

import pyperclip
from textual import on, work
from textual.widgets import TextArea


class SupportsClipboard(Protocol):
    """Minimal protocol for app clipboard support."""

    def copy_to_clipboard(self, text: str) -> None: ...


class CopyOnSelectTextArea(TextArea):
    """TextArea that automatically copies selected text to clipboard.

    This widget extends TextArea to provide copy-on-select functionality.
    When the user selects text, it is automatically copied to the clipboard
    after a short debounce period (to avoid copying while the user is still
    making their selection).

    The widget uses pyperclip for clipboard access with a fallback to
    OSC 52 escape sequences for terminals that support it, and pbcopy
    on macOS as a final fallback.

    Attributes:
        show_copy_notification: Whether to show a notification when text is copied.
            Defaults to False to avoid notification spam.
    """

    show_copy_notification: bool = False

    def __init__(
        self,
        *args: Any,
        show_copy_notification: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize the copy-on-select TextArea.

        Args:
            *args: Positional arguments passed to TextArea.
            show_copy_notification: Whether to show notifications when copying.
            **kwargs: Keyword arguments passed to TextArea.
        """
        super().__init__(*args, **kwargs)
        self._last_selection = ""
        self._pending_copy: str | None = None
        self.show_copy_notification = show_copy_notification

    @on(TextArea.SelectionChanged)
    def _on_selection_changed(self, event: TextArea.SelectionChanged) -> None:
        """Handle selection changes and schedule copy operation."""
        if event.selection.is_empty:
            self._pending_copy = None
            return

        selected = event.text_area.selected_text
        if selected and selected != self._last_selection:
            self._pending_copy = selected
            self._schedule_copy()

    @work(exclusive=True)
    async def _schedule_copy(self) -> None:
        """Wait for selection to stabilize before copying.

        This debounces the copy operation to avoid copying while the user
        is still making their selection. Waits 150ms after the last
        selection change before actually copying.
        """
        await asyncio.sleep(0.15)
        if self._pending_copy:
            self._copy_to_clipboard(self._pending_copy)
            self._last_selection = self._pending_copy

    def _copy_to_clipboard(self, text: str) -> None:
        """Copy text to clipboard using the best available method.

        Tries pyperclip first, then falls back to OSC 52 or pbcopy
        depending on the terminal environment.
        """
        try:
            pyperclip.copy(text)
            if self.show_copy_notification:
                char_count = len(text)
                label = f"{char_count} char{'s' if char_count != 1 else ''}"
                self.notify(f"Copied {label}", timeout=1)
        except Exception:
            # Fallback to OSC 52 or pbcopy
            self._copy_fallback(text)

    def _copy_fallback(self, text: str) -> None:
        """Fallback clipboard copy methods."""
        term_program = os.environ.get("TERM_PROGRAM", "")

        # Apple Terminal doesn't support OSC 52
        if term_program == "Apple_Terminal":
            self._copy_via_pbcopy(text)
        else:
            # Try OSC 52 (works over SSH in supported terminals)
            try:
                app_obj = getattr(self, "app", None)
                if app_obj is None:
                    raise RuntimeError("App not available")
                app = cast(SupportsClipboard, app_obj)
                app.copy_to_clipboard(text)
            except Exception:
                self._copy_via_pbcopy(text)

    def _copy_via_pbcopy(self, text: str) -> None:
        """Copy text using macOS pbcopy."""
        try:
            subprocess.run("pbcopy", text=True, input=text, check=True)
        except subprocess.SubprocessError:
            pass  # Silent failure - clipboard unavailable


class SelectableTextArea(CopyOnSelectTextArea):
    """Read-only TextArea for displaying selectable content.

    This widget provides a read-only text area that users can select text from.
    Selected text is automatically copied to the clipboard. This is useful for
    displaying content like chat messages or preview panes where users need
    to be able to copy specific portions of text.

    The widget is styled to look like a Static widget but with selection
    capabilities.

    Attributes:
        show_line_numbers: Always False for this widget.
        read_only: Always True for this widget.
    """

    DEFAULT_CSS = """
    SelectableTextArea {
        background: transparent;
        border: none;
        padding: 0;
        min-height: 1;
    }

    SelectableTextArea:focus {
        border: none;
    }

    SelectableTextArea > .text-area--cursor-line {
        background: transparent;
    }
    """

    def __init__(
        self,
        text: str = "",
        *,
        id: str | None = None,
        classes: str | None = None,
        show_copy_notification: bool = False,
    ) -> None:
        """Initialize the selectable text area.

        Args:
            text: Initial text content to display.
            id: The CSS ID of the widget.
            classes: CSS classes for the widget.
            show_copy_notification: Whether to show notifications when copying.
        """
        super().__init__(
            text,
            id=id,
            classes=classes,
            read_only=True,
            show_line_numbers=False,
            show_copy_notification=show_copy_notification,
        )

    def set_content(self, content: str) -> None:
        """Set the text content of the widget.

        This is a convenience method that replaces all text in the widget.

        Args:
            content: The new text content to display.
        """
        self.text = content



"""Shared flag validation and parsing utilities.

This module provides reusable validation functions for common flag patterns.
Import these in command handlers to avoid duplicating validation logic.

Usage:
    from cli.commands.flag_utils import validate_provider, validate_uuid, parse_tags

    def my_command(provider: str, id: str, tags: str | None) -> None:
        provider_enum = validate_provider(provider)
        uuid = validate_uuid(id, "resource")
        tag_list = parse_tags(tags)
"""

from uuid import UUID

import typer

from alloy_runtime_types.dtos.generation import REASONING_EFFORT_VALUES, ReasoningEffort
from alloy_runtime_types.dtos.knowledge import (
    SectionStrategyType,
    SourceType,
)
from alloy_runtime_types.enums.ownership_scope import OwnershipScope
from alloy_runtime_types.enums.provider import Provider
from alloy_runtime_types.enums.schema_enums import SchemaFormat
from alloy_runtime_types.enums.template_enums import (
    TemplateContentType,
    TemplateVisibility,
)


def validate_provider(provider: str) -> Provider:
    """Validate and convert provider string to Provider enum.

    Args:
        provider: Provider key string (e.g., 'openai', 'anthropic')

    Returns:
        Provider enum value

    Raises:
        typer.BadParameter: If provider is invalid
    """
    try:
        return Provider(provider.lower())
    except ValueError:
        valid = [p.value for p in Provider if not p.name.startswith("TESTING")]
        raise typer.BadParameter(
            f"Invalid provider '{provider}'. Valid: {', '.join(valid)}"
        )


def validate_uuid(value: str, name: str = "ID") -> UUID:
    """Validate and convert string to UUID.

    Args:
        value: UUID string
        name: Field name for error message (e.g., "agent", "template")

    Returns:
        UUID object

    Raises:
        typer.BadParameter: If UUID format is invalid
    """
    try:
        return UUID(value)
    except ValueError:
        raise typer.BadParameter(f"Invalid {name} UUID format: '{value}'")


def validate_scope(scope: str) -> OwnershipScope:
    """Validate and convert scope string to OwnershipScope enum.

    Args:
        scope: Scope string ('user', 'organization', 'global')

    Returns:
        OwnershipScope enum value

    Raises:
        typer.BadParameter: If scope is invalid
    """
    try:
        return OwnershipScope(scope.lower())
    except ValueError:
        valid = ", ".join(s.value for s in OwnershipScope)
        raise typer.BadParameter(f"Invalid scope '{scope}'. Valid: {valid}")


def validate_visibility(visibility: str) -> TemplateVisibility:
    """Validate and convert visibility string to TemplateVisibility enum.

    Args:
        visibility: Visibility string ('private', 'organization', 'public')

    Returns:
        TemplateVisibility enum value

    Raises:
        typer.BadParameter: If visibility is invalid
    """
    try:
        return TemplateVisibility(visibility.lower())
    except ValueError:
        valid = ", ".join(v.value for v in TemplateVisibility)
        raise typer.BadParameter(f"Invalid visibility '{visibility}'. Valid: {valid}")


def validate_template_content_type(content_type: str) -> TemplateContentType:
    """Validate and convert content type string to TemplateContentType enum.

    Args:
        content_type: Content type string ('system_instruction', 'user_message', 'fragment')

    Returns:
        TemplateContentType enum value

    Raises:
        typer.BadParameter: If content type is invalid
    """
    try:
        return TemplateContentType(content_type.lower())
    except ValueError:
        valid = ", ".join(ct.value for ct in TemplateContentType)
        raise typer.BadParameter(f"Invalid type '{content_type}'. Valid: {valid}")


def validate_schema_format(schema_format: str) -> SchemaFormat:
    """Validate and convert schema format string to SchemaFormat enum.

    Args:
        schema_format: Schema format string ('json_schema', 'regex')

    Returns:
        SchemaFormat enum value

    Raises:
        typer.BadParameter: If schema format is invalid
    """
    try:
        return SchemaFormat(schema_format.lower())
    except ValueError:
        valid = ", ".join(f.value for f in SchemaFormat)
        raise typer.BadParameter(
            f"Invalid schema format '{schema_format}'. Valid: {valid}"
        )


def validate_reasoning_effort(reasoning_effort: str) -> ReasoningEffort:
    lower = reasoning_effort.lower()
    if lower not in REASONING_EFFORT_VALUES:
        valid = ", ".join(REASONING_EFFORT_VALUES)
        raise typer.BadParameter(
            f"Invalid reasoning effort '{reasoning_effort}'. Valid: {valid}"
        )
    return lower


def validate_section_strategy(strategy: str) -> SectionStrategyType:
    """Validate section strategy for knowledge collection chunking.

    Args:
        strategy: Strategy string ('heading_based', 'paragraph_based', 'fixed_size', 'none')

    Returns:
        Valid SectionStrategyType value

    Raises:
        typer.BadParameter: If strategy is invalid
    """
    valid_strategies = ["heading_based", "paragraph_based", "fixed_size", "none"]
    lower = strategy.lower()
    if lower not in valid_strategies:
        raise typer.BadParameter(
            f"Invalid section strategy '{strategy}'. Valid: {', '.join(valid_strategies)}"
        )
    return lower  # type: ignore[return-value]


def validate_source_type(source_type: str) -> SourceType:
    """Validate source type for document ingestion.

    Args:
        source_type: Source type string ('upload', 'paste', 'api_import', 'url_scrape')

    Returns:
        Valid SourceType value

    Raises:
        typer.BadParameter: If source type is invalid
    """
    valid_types = ["upload", "paste", "api_import", "url_scrape"]
    lower = source_type.lower()
    if lower not in valid_types:
        raise typer.BadParameter(
            f"Invalid source type '{source_type}'. Valid: {', '.join(valid_types)}"
        )
    return lower  # type: ignore[return-value]


def validate_document_type(
    document_type: str,
) -> str:
    """Validate document type for preprocessing.

    Args:
        document_type: Document type string ('transcript', 'markdown', 'document', 'form')

    Returns:
        Valid document type string

    Raises:
        typer.BadParameter: If document type is invalid
    """
    valid_types = ["transcript", "markdown", "document", "form"]
    lower = document_type.lower()
    if lower not in valid_types:
        raise typer.BadParameter(
            f"Invalid document type '{document_type}'. Valid: {', '.join(valid_types)}"
        )
    return lower


def validate_output_mode(
    output_mode: str,
    output_file: str | None,
    valid_modes: tuple[str, ...] = ("console", "clipboard", "file"),
) -> None:
    """Validate output mode and file path combination.

    Args:
        output_mode: Output mode ('console', 'clipboard', 'file', 'editor')
        output_file: Output file path (required for 'file' mode)
        valid_modes: Tuple of valid mode strings

    Raises:
        typer.BadParameter: If mode is invalid or file required but not provided
    """
    if output_mode not in valid_modes:
        raise typer.BadParameter(
            f"Invalid output '{output_mode}'. Valid: {', '.join(valid_modes)}"
        )
    if output_mode == "file" and not output_file:
        raise typer.BadParameter("--output-file/-O required with --output file")


def parse_tags(tags: str | None) -> list[str]:
    """Parse comma-separated tags string into list.

    Args:
        tags: Comma-separated tags string or None

    Returns:
        List of tag strings (empty list if None)

    Example:
        parse_tags("email,marketing,draft") -> ["email", "marketing", "draft"]
        parse_tags(None) -> []
    """
    if not tags:
        return []
    return [t.strip() for t in tags.split(",") if t.strip()]


def parse_tags_required(tags: str) -> list[str]:
    """Parse comma-separated tags string, requiring at least one tag.

    Args:
        tags: Comma-separated tags string

    Returns:
        List of tag strings

    Raises:
        typer.BadParameter: If no valid tags after parsing
    """
    tag_list = parse_tags(tags)
    if not tag_list:
        raise typer.BadParameter("At least one tag required")
    return tag_list


def require_one_of(*args: tuple[str, object]) -> None:
    """Require that at least one of the named values is provided.

    Args:
        *args: Tuples of (name, value) to check

    Raises:
        typer.BadParameter: If all values are None/empty

    Example:
        require_one_of(
            ("--message", message),
            ("--message-template", message_template),
        )
    """
    provided = [name for name, value in args if value]
    if not provided:
        names = " or ".join(name for name, _ in args)
        raise typer.BadParameter(f"One of {names} is required")


def require_exclusive(*args: tuple[str, object]) -> None:
    """Require that at most one of the named values is provided.

    Args:
        *args: Tuples of (name, value) to check

    Raises:
        typer.BadParameter: If more than one value is provided

    Example:
        require_exclusive(
            ("--agent", agent),
            ("--provider/--model", provider or model),
        )
    """
    provided = [name for name, value in args if value]
    if len(provided) > 1:
        raise typer.BadParameter(f"Cannot use {' and '.join(provided)} together")


def require_together(*args: tuple[str, object]) -> None:
    """Require that if any of the named values is provided, all must be provided.

    Args:
        *args: Tuples of (name, value) to check

    Raises:
        typer.BadParameter: If some but not all values are provided

    Example:
        require_together(
            ("--provider", provider),
            ("--model", model),
        )
    """
    provided = [name for name, value in args if value]
    missing = [name for name, value in args if not value]

    if provided and missing:
        raise typer.BadParameter(
            f"{' and '.join(missing)} required with {' and '.join(provided)}"
        )


def check_no_update_flags(*args: tuple[str, object]) -> bool:
    """Check if no update flags were provided.

    Args:
        *args: Tuples of (name, value) to check

    Returns:
        True if no flags provided

    Example:
        if check_no_update_flags(("--name", name), ("--description", desc)):
            raise typer.BadParameter("At least one update flag required")
    """
    return all(value is None for _, value in args)


def require_update_flags(*args: tuple[str, object]) -> None:
    """Require that at least one update flag was provided.

    Args:
        *args: Tuples of (name, value) to check

    Raises:
        typer.BadParameter: If no flags provided
    """
    if check_no_update_flags(*args):
        names = ", ".join(name for name, _ in args)
        raise typer.BadParameter(f"At least one flag required: {names}")



"""Presenters for pipeline schedules command output."""

from datetime import datetime

from alloy_runtime_types.dtos.schedule import (
    DeleteScheduleResponse,
    GetScheduleResponse,
    ListSchedulesResponse,
)

from cli.infrastructure.output import OutputFormat, OutputService
from cli.infrastructure.renderers.list_renderer import ColumnConfig


def present_schedules_list(response: ListSchedulesResponse) -> None:
    """Present list of schedules with Rich table formatting."""
    output = OutputService.get()

    if getattr(output, "format", OutputFormat.COMPACT) == OutputFormat.JSON:
        output.raw(response)
        return

    columns = [
        ColumnConfig(
            source_field="id",
            header_label="ID",
            compact_line=1,
            compact_style="cyan",
        ),
        ColumnConfig(
            source_field="is_enabled",
            header_label="Enabled",
            formatter=lambda x: "Yes" if x else "No",
            compact_line=1,
            compact_style="green",
        ),
        ColumnConfig(
            source_field="schedule_type",
            header_label="Type",
            compact_line=1,
            compact_style="yellow",
        ),
        ColumnConfig(
            source_field="name",
            header_label="Name",
            compact_line=2,
            compact_style="white bold",
        ),
        ColumnConfig(
            source_field="next_run_at",
            header_label="Next Run",
            formatter=lambda x: x.strftime("%Y-%m-%d %H:%M") if x else "Never",
            compact_line=2,
            compact_style="blue",
        ),
    ]

    output.table(
        items=response.schedules,  # type: ignore[arg-type]
        title=f"Schedules ({response.total})",
        columns=columns,
        raw_payload=response,
        total_count=response.total,
    )


def present_schedule_detail(response: GetScheduleResponse) -> None:
    """Present schedule details with Rich formatting."""
    output = OutputService.get()

    from cli.infrastructure.formatting.fields import format_optional
    from cli.infrastructure.renderers.entity_renderer import FieldConfig

    def _format_datetime_with_tz(dt: datetime | None) -> str:
        """Format datetime with timezone."""
        return format_optional(dt, lambda x: x.strftime("%Y-%m-%d %H:%M:%S %Z"))

    fields = [
        FieldConfig("id", "Schedule ID", str),
        FieldConfig("pipeline_id", "Pipeline ID", str),
        FieldConfig("name", "Name"),
        FieldConfig("description", "Description", format_optional),
        FieldConfig("schedule_type", "Type"),
        FieldConfig("timezone", "Timezone"),
        FieldConfig("is_enabled", "Enabled", lambda v: "Yes" if v else "No"),
        FieldConfig("cron_expression", "Cron Expression", format_optional),
        FieldConfig("rrule_freq", "RRULE Freq", format_optional),
        FieldConfig(
            "rrule_interval", "RRULE Interval", lambda v: str(v) if v and v > 1 else "-"
        ),
        FieldConfig("next_run_at", "Next Run", _format_datetime_with_tz),
        FieldConfig("last_run_at", "Last Run", _format_datetime_with_tz),
        FieldConfig("consecutive_failures", "Consecutive Failures", str),
        FieldConfig(
            "allow_concurrent", "Allow Concurrent", lambda v: "Yes" if v else "No"
        ),
        FieldConfig("max_consecutive_failures", "Max Failures", str),
    ]

    output.entity(response, f"Schedule: {response.name}", fields, raw_payload=response)


def present_delete_result(response: DeleteScheduleResponse) -> None:
    """Present schedule deletion result."""
    output = OutputService.get()
    output.result(
        action="delete",
        resource="schedule",
        identifier=str(response.id),
        name=None,
    )



"""Presenter for content delete command output."""

from cli.infrastructure.output import OutputService
from alloy_runtime_types.dtos.content import DeleteContentPartResponse


def present_delete_result(response: DeleteContentPartResponse) -> None:
    """Present content part deletion result.

    Args:
        response: DeleteContentPartResponse from server
    """
    output = OutputService.get()
    deleted = response.deleted_content_part

    output.result(
        action="delete",
        resource="content_part",
        identifier=str(deleted.id),
        name=None,
    )



"""Tool config creation modal for creating new tool configurations.

Provides a modal screen that allows users to create a new tool configuration
with required fields: name, tool_id, config JSON, and optional description.
"""

import json
from dataclasses import dataclass
from typing import Any

from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.widgets import (
    Button,
    Input,
    Label,
    RadioButton,
    RadioSet,
    Static,
    TextArea,
)

from cli.infrastructure.scope_utils import get_scope_from_radioset
from cli.tui.widgets.base_form_modal import BaseFormModal

from alloy_runtime_sdk.api_client.client import ApiClient
from alloy_runtime_types.dtos.tool_configs import CreateToolConfigRequest
from alloy_runtime_types.enums.ownership_scope import OwnershipScope


@dataclass
class ToolConfigCreateResult:
    """Result of successful tool config creation."""

    config_id: str
    config_name: str


class ToolConfigCreateModal(BaseFormModal[ToolConfigCreateResult]):
    """Modal for creating a new tool configuration.

    Provides a form with:
    - Required fields: name, tool_id, config (JSON)
    - Scope selection (user/organization/global)
    - Optional: description

    Usage:
        def on_config_created(result: ToolConfigCreateResult | None):
            if result is not None:
                self.notify(f"Created config: {result.config_name}")
                self._do_search("")  # Refresh list

        client = await self.app.store.get_client()
        self.app.push_screen(
            ToolConfigCreateModal(client=client),
            on_config_created
        )
    """

    def __init__(
        self,
        client: ApiClient,
        **kwargs: Any,
    ) -> None:
        """Initialize the tool config creation modal.

        Args:
            client: ApiClient instance for API calls
        """
        super().__init__(**kwargs)
        self._client = client

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="tool-config-create-modal"):
            yield Label("Create Tool Config", classes="modal-title")

            # ===== REQUIRED FIELDS =====
            yield Static("[bold]Required[/]", classes="form-section")

            yield Static("Name", classes="field-label")
            yield Input(
                placeholder="e.g., RAG with HyDE, Fast Web Search",
                id="name-input",
            )

            yield Static("Tool ID", classes="field-label")
            yield Input(
                placeholder="e.g., rag_search, web_search",
                id="tool-id-input",
            )

            yield Static("Scope", classes="field-label")
            with RadioSet(id="scope-radios"):
                yield RadioButton("User (personal)", id="scope-user")
                yield RadioButton("Organization", id="scope-org", value=True)
                yield RadioButton("Global", id="scope-global")

            yield Static(
                "Configuration (JSON object with tool parameters)",
                classes="field-label",
            )
            yield TextArea(
                id="config-textarea",
                classes="definition-input",
            )

            # ===== OPTIONAL FIELDS =====
            yield Static("[bold]Optional[/]", classes="form-section")

            yield Static("Description", classes="field-label")
            yield Input(
                placeholder="Brief description of this configuration",
                id="description-input",
            )

            # Error display
            yield Static("", id="error-display", classes="error-display")

            with Horizontal(classes="button-row"):
                yield Button("Create Config", id="create-btn", variant="success")
                yield Button("Cancel", id="cancel-btn", variant="default")

    def on_mount(self) -> None:
        """Focus the name input and set default config on mount."""
        self.query_one("#config-textarea", TextArea).text = "{}"
        self.query_one("#name-input", Input).focus()

    @on(Button.Pressed, "#create-btn")
    def on_create_pressed(self, event: Button.Pressed) -> None:
        """Handle create button press."""
        self.action_submit()

    @on(Button.Pressed, "#cancel-btn")
    def on_cancel_pressed(self, event: Button.Pressed) -> None:
        """Handle cancel button press."""
        self.dismiss(None)

    def action_submit(self) -> None:
        """Handle form submission."""
        try:
            self._validate_fields()
            self._hide_error()
            self._finalize_submission()
        except ValueError as e:
            self._show_error(str(e))

    def _validate_fields(self) -> None:
        """Validate all required fields."""
        name = self.query_one("#name-input", Input).value.strip()
        if not name:
            raise ValueError("Name is required")
        if len(name) > 255:
            raise ValueError("Name must be 255 characters or less")

        tool_id = self.query_one("#tool-id-input", Input).value.strip()
        if not tool_id:
            raise ValueError("Tool ID is required")

        config_text = self.query_one("#config-textarea", TextArea).text.strip()
        if not config_text:
            raise ValueError("Configuration is required")

        try:
            config = json.loads(config_text)
            if not isinstance(config, dict):
                raise ValueError(
                    "Configuration must be a JSON object (not array/primitive)"
                )
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in configuration: {e.msg}")

    def _get_scope(self) -> OwnershipScope:
        """Get scope from RadioSet."""
        return get_scope_from_radioset(self.query_one("#scope-radios", RadioSet))

    def _build_request(self) -> CreateToolConfigRequest:
        """Build the API request."""
        name = self.query_one("#name-input", Input).value.strip()
        tool_id = self.query_one("#tool-id-input", Input).value.strip()
        config_text = self.query_one("#config-textarea", TextArea).text.strip()
        config = json.loads(config_text)
        description = self.query_one("#description-input", Input).value.strip() or None

        return CreateToolConfigRequest(
            name=name,
            tool_id=tool_id,
            config=config,
            description=description,
            scope=self._get_scope(),
        )

    @work(exclusive=True)
    async def _finalize_submission(self) -> None:
        """Submit to API."""
        try:
            request = self._build_request()
            response = await self._client.create_tool_config(request)

            result = ToolConfigCreateResult(
                config_id=str(response.id),
                config_name=response.name,
            )
            self.dismiss(result)

        except Exception as e:
            error_msg = str(e)
            # Provide user-friendly error messages
            if "409" in error_msg or "conflict" in error_msg.lower():
                self._show_error(
                    "A tool config with this name already exists in this scope"
                )
            elif "validation" in error_msg.lower():
                self._show_error(f"Invalid configuration: {error_msg}")
            else:
                self._show_error(f"Failed to create tool config: {error_msg}")



"""Presenter for content edit command output."""

import json

from cli.infrastructure.formatting.fields import (
    format_cost,
    format_datetime,
    format_optional,
    format_tokens,
    format_uuid,
)
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.entity_renderer import FieldConfig
from alloy_runtime_types.dtos.content import (
    ExecutionMetadataResponse,
    UpdateContentPartResponse,
)
from alloy_runtime_types.dtos.templates import TagResponse


def present_content_edit_success(
    response: UpdateContentPartResponse, updated_content: bool, updated_tags: bool
) -> None:
    output = OutputService.get()
    execution_metadata = _require_assistant_execution_metadata(response)

    updates: list[str] = []
    if updated_content:
        updates.append("content")
    if updated_tags:
        updates.append("tags")
    update_summary = " and ".join(updates) if updates else "no changes"

    output.result(
        action="update",
        resource="content_part",
        status="success",
        identifier=str(response.id),
        extra_fields={"Updated": update_summary},
    )

    metadata_fields = [
        FieldConfig("id", "ID", str),
        FieldConfig(
            "message_id", "Message ID", lambda value: format_uuid(value, short=True)
        ),
        FieldConfig("part_index", "Part Index", str),
        FieldConfig("content_type_id", "Content Type"),
        FieldConfig(
            "content_structured",
            "Storage Type",
            lambda value: "structured" if value else "text",
        ),
        FieldConfig("created_at", "Created", format_datetime),
        FieldConfig("updated_at", "Updated At", format_datetime),
        FieldConfig("tags", "Tags", _format_tags),
        FieldConfig(
            "id",
            "Execution ID",
            lambda _: format_uuid(execution_metadata.agent_execution_id, short=True),
        ),
        FieldConfig(
            "id",
            "Agent",
            lambda _: format_optional(
                execution_metadata.agent_name,
                placeholder="(direct model call)",
            ),
        ),
        FieldConfig("id", "Provider", lambda _: execution_metadata.provider_key),
        FieldConfig("id", "Model", lambda _: execution_metadata.provider_model_name),
        FieldConfig(
            "id",
            "Cost",
            lambda _: format_optional(
                execution_metadata.total_cost_usd,
                formatter=format_cost,
            ),
        ),
        FieldConfig(
            "id",
            "Tokens",
            lambda _: format_optional(
                execution_metadata.total_tokens,
                formatter=format_tokens,
            ),
        ),
        FieldConfig(
            "id",
            "Queued At",
            lambda _: format_datetime(execution_metadata.queued_at),
        ),
    ]
    output.entity(
        response, "Updated Content Part", metadata_fields, raw_payload=response
    )

    content_fields = [FieldConfig("id", "Content", lambda _: _format_content(response))]
    output.entity(response, "Updated Content Body", content_fields)


def _format_tags(tags: list[TagResponse]) -> str:
    """Format tags list for display.

    Args:
        tags: List of tag responses

    Returns:
        Formatted tags string
    """
    if not tags:
        return "(no tags)"

    tag_paths = [t.tag_path for t in tags]
    return ", ".join(tag_paths)


def _format_content(response: UpdateContentPartResponse) -> str:
    if response.content_text is not None:
        return response.content_text
    if response.content_structured is not None:
        return json.dumps(response.content_structured, indent=2, ensure_ascii=False)
    return "(no content)"


def _require_assistant_execution_metadata(
    response: UpdateContentPartResponse,
) -> ExecutionMetadataResponse:
    assert response.message_role == "assistant"
    execution_metadata = response.execution_metadata
    assert execution_metadata is not None
    return execution_metadata



"""Knowledge synthesis update command implementation."""

from typing import Optional

import typer

from cli.commands.knowledge.synthesis.update.presenter import (
    present_synthesis_update_result,
)
from cli.infrastructure.command import async_command, authenticated_client
from cli.infrastructure.file_content import resolve_content_or_raise
from alloy_runtime_types.dtos.knowledge import UpdateSynthesisRequest


def knowledge_synthesis_update_command(
    synthesis_id: str = typer.Argument(
        ...,
        help="Synthesis UUID",
    ),
    title: Optional[str] = typer.Option(
        None,
        "--title",
        help="New title for the synthesis",
    ),
    content: Optional[str] = typer.Option(
        None,
        "--content",
        help="New synthesis content (or @file)",
    ),
) -> None:
    """Update a synthesis document.

    Update synthesis properties like title or content.
    At least one of --title or --content must be provided.

    The --content flag accepts a text string or @file reference.

    Examples:
        alloy knowledge synthesis update <uuid> --title "New Title"
        alloy knowledge synthesis update <uuid> --content @updated.md
        alloy knowledge synthesis update <uuid> --title "New Title" --content "Updated content"
    """
    _execute_update(
        synthesis_id=synthesis_id,
        title=title,
        content_raw=content,
    )


@async_command
async def _execute_update(
    synthesis_id: str,
    title: Optional[str],
    content_raw: Optional[str],
) -> None:
    """Execute synthesis update operation."""

    resolved_content = resolve_content_or_raise(content_raw)

    if title is None and resolved_content is None:
        raise typer.BadParameter(
            "At least one of --title or --content must be provided"
        )

    request = UpdateSynthesisRequest(
        title=title,
        synthesis_content=resolved_content,
    )

    async with authenticated_client() as (_config, client):
        response = await client.update_synthesis(
            synthesis_id=synthesis_id,
            request=request,
        )

    present_synthesis_update_result(response)



"""Presenter for schemas list command output."""

from pydantic import BaseModel, Field
from alloy_runtime_types.dtos.schemas import ListSchemasResponse, SchemaSummary

from cli.infrastructure.scope_utils import format_scope_label
from cli.infrastructure.output import OutputService
from cli.infrastructure.renderers.list_renderer import ColumnConfig


class SchemaDisplay(BaseModel):
    name: str = Field(..., description="Schema name")
    id: str = Field(..., description="Schema ID")
    version: int = Field(..., description="Version number")
    format: str = Field(..., description="Format type")
    scope: str = Field(..., description="Ownership scope")
    created: str | None = Field(None, description="Creation date")
    description: str | None = Field(None, description="Description")


def _transform_schema_for_display(schema: SchemaSummary) -> SchemaDisplay:
    return SchemaDisplay(
        name=schema.schema_name,
        id=str(schema.id),
        version=schema.version,
        format=schema.schema_format_id,
        scope=format_scope_label(
            str(schema.organization_id) if schema.organization_id else None,
            str(schema.user_id) if schema.user_id else None,
        ),
        created=schema.created_at.strftime("%Y-%m-%d") if schema.created_at else None,
        description=schema.description,
    )


def present_schemas_list(response: ListSchemasResponse) -> None:
    output = OutputService.get()
    display_items = [_transform_schema_for_display(s) for s in response.schemas]
    columns = [
        ColumnConfig(
            source_field="name",
            header_label="Name",
        ),
        ColumnConfig(
            source_field="id",
            header_label="ID",
        ),
        ColumnConfig(
            source_field="version",
            header_label="Version",
        ),
        ColumnConfig(
            source_field="format",
            header_label="Format",
        ),
        ColumnConfig(
            source_field="scope",
            header_label="Scope",
        ),
        ColumnConfig(
            source_field="created",
            header_label="Created",
        ),
        ColumnConfig(
            source_field="description",
            header_label="Description",
        ),
    ]

    output.table(
        items=display_items,  # type: ignore[arg-type]
        title="Schemas",
        columns=columns,
        raw_payload=response,
    )
